
'''
Decode "candump" CAN frames or the output of "cantools decode"
read from standard input and plot them using matplotlib.
You can select which signals to plot by specifying them on the command line.
Each signal is one argument and has the pattern "[bo.]sg[:fmt]"
where bo is the name of the message, sg is the name of the signal
and fmt is the format of the graph.
The wildcards * (any number of any character)
and ? (exactly one arbitrary character)
can be used inside of sg and bo.
If bo is omitted it defaults to *.

fmt is passed to matplotlib and can be used to specify
the line style, markers and color.
For example the following values can be combined:
Line style:
    '-'  solid line style,
    '--' dashed line style,
    '-.' dash-dot line style and
    ':'  dotted line style.
Markers:
    '.' point marker,
    ',' pixel marker,
    'o' circle marker,
    's' square marker,
    'D' diamond marker,
    'x' x marker
    and many more.
Colors:
    'b' blue,
    'g' green,
    'r' red,
    'c' cyan,
    'm' magenta,
    'y' yellow,
    'k' black and
    'w' white.
    'C0'...'C9' the colors defined by the current style
https://matplotlib.org/api/_as_gen/matplotlib.pyplot.plot.html

If the first character of fmt is a '|' stem is used instead of plot.

Signals can be separated by a '-' to show them in different subplots.

Signals can be separated by a ',' to make them refer to different vertical axes in the same subplot.
I recommend using this with the option --auto-color-ylabels.

All signals (independent of the subplot and vertical axis) share the same horizontal axis.
'''

import argparse
import binascii
import datetime
import re
import struct
import sys

from argparse_addons import Integer  # type: ignore

try:
    from matplotlib import pyplot as plt
except ImportError:
    plt = None  # type: ignore[assignment,unused-ignore]

from .. import database, errors
from ..database.namedsignalvalue import NamedSignalValue

PYPLOT_BASE_COLORS = "bgrcmykwC"


class MatplotlibNotInstalledError(errors.Error):

    def __init__(self):
        super().__init__("The matplotlib package not installed and is required "
                         "for producing plots.")


if plt is not None:
    #TODO: I am not allowing "%H:%M" as input (for --start or --stop) because it could be misinterpreted as "%M:%S". Should this output format be changed?
    # I don't think the ambiguity is a problem for the output because if it is not obvious from the context it can be easily clarified with --xlabel.
    # However, it seems very unintuitive if the same format which is used for output is not allowed for input.
    # If you do change it, remember to uncomment the tests in test_plot_unittests.py.
    plt.rcParams["date.autoformatter.hour"] = "%H:%M"
    plt.rcParams["date.autoformatter.minute"] = "%H:%M"
    plt.rcParams["date.autoformatter.microsecond"] = "%H:%M:%S.%f"


# Matches 'candump' output, i.e. "vcan0  1F0   [8]  00 00 00 00 00 00 1B C1".
RE_CANDUMP = re.compile(r'^\s*(?:\((?P<time>.*?)\))?\s*\S+\s+(?P<frameid>[0-9A-F]+)\s*\[\d+\]\s*(?P<data>[0-9A-F ]*)(?:\s*::.*)?$')
# Matches 'cantools decode' output, i.e. ")" or "   voltage: 0 V,".
RE_DECODE = re.compile(r'\w+\(|\s+\w+:\s+[0-9.+-]+(\s+.*)?,?|\)')
# Matches 'candump -l' (or -L) output, i.e. "(1594172461.968006) vcan0 1F0#0000000000001BC1"
RE_CANDUMP_LOG = re.compile(r'^\((?P<time>\d+\.\d+)\)\s+\S+\s+(?P<frameid>[\dA-F]+)#(?P<data>[\dA-F]*)(\s+[RT])?$')


def _mo_unpack(mo):
    '''extract the data from a re match object'''
    timestamp = mo.group('time')
    frame_id = mo.group('frameid')
    frame_id = '0' * (8 - len(frame_id)) + frame_id
    frame_id = binascii.unhexlify(frame_id)
    frame_id = struct.unpack('>I', frame_id)[0]
    data = mo.group('data')
    data = data.replace(' ', '')
    data = binascii.unhexlify(data)

    return timestamp, frame_id, data

class TimestampParser:

    '''
    Parses the values for the horizontal axis
    and generates the corresponding axis label.
    Preferably timestamps are used but if none
    are given it falls back to line numbers.
    '''

    # candump -ta, -tz and -td have the same timestamp syntax: a floating number in seconds.
    # In case of -td using timestamps does not seem useful and a user should use --line-numbers.
    # The following constant shall distinguish between -ta and -tz.
    # If the first timestamp is bigger than THRESHOLD_ABSOLUTE_SECONDS I am assuming -ta is used
    # and convert timestamps to datetime objects which will print a date.
    # Otherwise I'll assume -tz is used and format them using timedelta objects.
    # I am not using zero to compare against in case the beginning of the log file is stripped.
    THRESHOLD_ABSOLUTE_SECONDS = 60*60*24*7

    FORMAT_ABSOLUTE_TIMESTAMP = "%Y-%m-%d %H:%M:%S.%f"

    def __init__(self, args):
        self.use_timestamp = None
        self.relative = None
        self._parse_timestamp = None
        self.first_timestamp = None
        self.args = args

    def init_start_stop(self, x0):
        if self.use_timestamp and self.relative:
            parse = self.parse_user_input_relative_time
        elif self.use_timestamp:
            parse = self.parse_user_input_absolute_time
        else:
            def parse(s, _x0):
                return int(s)

        if self.args.start is not None:
            self.args.start = parse(self.args.start, x0)
            x0 = self.args.start
            self.first_timestamp = x0
        if self.args.stop is not None:
            self.args.stop = parse(self.args.stop, x0)

    def parse_user_input_relative_time(self, user_input, first_timestamp):
        try:
            return float(user_input)
        except ValueError:
            pass

        patterns_hour = ['%H:%M:', '%H:%M:%S', '%H:%M:%S.%f']
        patterns_minute = [':%M:%S', '%M:%S.', '%M:%S.%f']
        patterns_day = ['%d day', '%d days']

        day_time_sep = ', '
        for pattern_day in tuple(patterns_day):
            for pattern_time in ['%H:%M', *patterns_hour]:
                patterns_day.append(pattern_day+day_time_sep+pattern_time)

        for pattern in patterns_minute + patterns_hour + patterns_day:
            t = self.strptimedelta_in_seconds(user_input, pattern)
            if t is not None:
                return t

        raise ValueError(f"Failed to parse relative time {user_input!r}.\n\nPlease note that an input like 'xx:xx' is ambiguous. It could be either 'HH:MM' or 'MM:SS'. Please specify what you want by adding a leading or trailing colon: 'HH:MM:' or ':MM:SS' (or 'MM:SS.').")

    def strptimedelta_in_seconds(self, user_input, pattern):
        '''
        Parse the string representation of a time delta object.
        Return value: int in seconds or None if parsing failed.
        '''
        # I cannot use `datetime.datetime.strptime(user_input, pattern) - datetime.datetime.strptime("", "")` because it treats no day as 1 day
        p = pattern
        p = p.replace('%H', '{hour}')
        p = p.replace('%M', '{min}')
        p = p.replace('%S', '{s}')
        p = p.replace('%f', '{ms}')
        p = p.replace('%d', '{day}')
        p = re.escape(p)
        p = p.replace(r'\{hour\}', '(?P<hour>[0-9][0-9]?)')
        p = p.replace(r'\{min\}', '(?P<min>[0-9][0-9]?)')
        p = p.replace(r'\{s\}', '(?P<s>[0-9][0-9]?)')
        p = p.replace(r'\{ms\}', '(?P<ms>[0-9]+)')
        p = p.replace(r'\{day\}', '(?P<day>[0-9][0-9]?)')
        p += '$'
        m = re.match(p, user_input)
        if m is None:
            return None

        d = m.groupdict('0')
        seconds = float(d.pop('s','0') + '.' + d.pop('ms','0'))
        d = {key:int(d[key]) for key in d}
        return ((d.pop('day',0)*24 + d.pop('hour',0))*60 + d.pop('min',0))*60 + seconds

    def parse_user_input_absolute_time(self, user_input, first_timestamp):
        patterns_year = ['%Y-%m-%d', '%d.%m.%Y']
        patterns_month = ['%m-%d', '%d.%m.']
        patterns_day = ['%d.']
        patterns_hour = ['%H:%M:', '%H:%M:%S', '%H:%M:%S.%f']
        patterns_minute = [':%M:%S', '%M:%S.', '%M:%S.%f']
        patterns_second = ['%S', '%S.%f']

        date_time_sep = ' '
        for patterns in (patterns_year, patterns_month, patterns_day):
            for pattern_date in tuple(patterns):
                for pattern_time in ['%H:%M', *patterns_hour]:
                    patterns.append(pattern_date+date_time_sep+pattern_time)

        patterns_year.append('%Y-%m')

        for attrs, patterns in [
            (['year', 'month', 'day', 'hour', 'minute'], patterns_second),
            (['year', 'month', 'day', 'hour'], patterns_minute),
            (['year', 'month', 'day'], patterns_hour),
            (['year', 'month'], patterns_day),
            (['year'], patterns_month),
            ([], patterns_year),
        ]:
            for p in patterns:
                try:
                    # HACK: ensure that an absolute year is specified
                    # to make it work with python >= 3.15 (and get
                    # around the deprecation warning for earlier
                    # versions). We use 1904 because it is a leap year
                    # and because it should be clear that this is not
                    # the real year of the specified date.
                    if '%Y' not in p and '%d' in p:
                        out = datetime.datetime.strptime('1904 ' + user_input, '%Y ' + p)  # noqa: DTZ007
                    else:
                        out = datetime.datetime.strptime(user_input, p)  # noqa: DTZ007
                except ValueError:
                    pass
                else:
                    kw = {a:getattr(first_timestamp,a) for a in attrs}
                    out = out.replace(**kw)
                    return out

        raise ValueError(f"Failed to parse absolute time {user_input!r}.\n\nPlease note that an input like 'xx:xx' is ambiguous. It could be either 'HH:MM' or 'MM:SS'. Please specify what you want by adding a leading or trailing colon: 'HH:MM:' or ':MM:SS' (or 'MM:SS.').")

    def first_parse_timestamp(self, timestamp, linenumber):
        if timestamp is None:
            self.use_timestamp = False
            return linenumber

        try:
            out = self.parse_absolute_timestamp(timestamp)
            self.use_timestamp = True
            self.relative = False
            self.first_timestamp = out
            self._parse_timestamp = self.parse_absolute_timestamp
            return out
        except ValueError:
            pass

        try:
            if float(timestamp) > self.THRESHOLD_ABSOLUTE_SECONDS:
                out = self.parse_absolute_seconds(timestamp)
                self.relative = False
                self.first_timestamp = out
                self._parse_timestamp = self.parse_absolute_seconds
            else:
                out = self.parse_seconds(timestamp)
                self.relative = True
                self._parse_timestamp = self.parse_seconds

            self.use_timestamp = True
            return out
        except ValueError:
            pass

        self.use_timestamp = False
        return linenumber

    def parse_timestamp(self, timestamp, linenumber):
        if self.use_timestamp is None:
            x = self.first_parse_timestamp(timestamp, linenumber)
            self.init_start_stop(x)
            return x

        if self.use_timestamp:
            return self._parse_timestamp(timestamp)
        else:
            return linenumber

    def parse_absolute_timestamp(self, timestamp):
        return datetime.datetime.strptime(timestamp, self.FORMAT_ABSOLUTE_TIMESTAMP)  # noqa: DTZ007

    @staticmethod
    def parse_absolute_seconds(timestamp):
        return datetime.datetime.fromtimestamp(float(timestamp))  # noqa: DTZ006

    @staticmethod
    def parse_seconds(timestamp):
        return float(timestamp)

    def get_label(self):
        if self.use_timestamp:
            if self.relative:
                label = "relative time"
            else:
                label = "absolute time"
        else:
            label = "line number"

        if isinstance(self.first_timestamp, datetime.datetime):
            label += self.first_timestamp.strftime(" (start: %d.%m.%Y)")

        return label

def _do_decode(args):
    '''
    The entry point of the program.
    It iterates over all input lines, parses them
    and passes the data to a Plotter object.
    '''
    if plt is None:
        raise MatplotlibNotInstalledError()

    if args.list_styles:
        print("available matplotlib styles:")
        for style in plt.style.available:
            print(f"- {style}")
        return

    if args.show_errors:
        args.show_invalid_syntax = True
        args.show_unknown_frames = True
        args.show_invalid_data = True
    if args.quiet:
        args.ignore_invalid_syntax = True
        args.ignore_unknown_frames = True
        args.ignore_invalid_data = True

    dbase = database.load_file(args.database,
                               encoding=args.encoding,
                               frame_id_mask=args.frame_id_mask,
                               prune_choices=args.prune,
                               strict=not args.no_strict)
    re_format = None
    timestamp_parser = TimestampParser(args)
    if args.show_invalid_syntax:
        # we cannot use a timestamp if we have failed to parse the line
        timestamp_parser.use_timestamp = False
    if args.line_numbers:
        timestamp_parser.use_timestamp = False

    if args.style is not None:
        plt.style.use(args.style)

    plotter = Plotter(dbase, args)

    line_number = 1
    while True:
        line = sys.stdin.readline()

        # Break at EOF.
        if not line:
            break

        line = line.strip('\r\n')
        if not line:
            continue

        # Auto-detect on first valid line.
        if re_format is None:
            mo = RE_CANDUMP.match(line)

            if mo:
                re_format = RE_CANDUMP
            else:
                mo = RE_CANDUMP_LOG.match(line)

                if mo:
                    re_format = RE_CANDUMP_LOG
        else:
            mo = re_format.match(line)

        if mo:
            timestamp, frame_id, data = _mo_unpack(mo)
            timestamp = timestamp_parser.parse_timestamp(timestamp, line_number)
            if args.start is not None and timestamp < args.start:
                line_number += 1
                continue
            elif args.stop is not None and timestamp > args.stop:
                break
            plotter.add_msg(timestamp, frame_id, data)
        elif RE_DECODE.match(line):
            continue
        else:
            plotter.failed_to_parse_line(line_number, line)

        line_number += 1

    plotter.plot(timestamp_parser.get_label())


class Plotter:

    '''
    Decodes the data received from _do_decode further
    and stores them in a Signals object.
    Shows or exports the data plotted by Signals.
    '''

    # ------- initialization -------

    def __init__(self, dbase, args):
        self.dbase = dbase
        self.decode_choices = not args.no_decode_choices
        self.show_invalid_syntax = args.show_invalid_syntax
        self.show_unknown_frames = args.show_unknown_frames
        self.show_invalid_data = args.show_invalid_data
        self.ignore_invalid_syntax = args.ignore_invalid_syntax
        self.ignore_unknown_frames = args.ignore_unknown_frames
        self.ignore_invalid_data = args.ignore_invalid_data
        self.output_filename = args.output_file
        self.signals = Signals(args.signals, args.case_sensitive, args.break_time, args, args.auto_color_ylabels, dbase)

        self.x_invalid_syntax = []
        self.x_unknown_frames = []
        self.x_invalid_data = []

    # ------- while reading data -------

    def add_msg(self, timestamp, frame_id, data):
        try:
            message = self.dbase.get_message_by_frame_id(frame_id)
        except KeyError:
            if self.show_unknown_frames:
                self.x_unknown_frames.append(timestamp)
            if not self.ignore_unknown_frames:
                print(f'Unknown frame id {frame_id} (0x{frame_id:x})')
            return

        try:
            decoded_signals = message.decode(data, self.decode_choices)
        except Exception as e:
            if self.show_invalid_data:
                self.x_invalid_data.append(timestamp)
            if not self.ignore_invalid_data:
                print(f'Failed to parse data of frame id {frame_id} (0x{frame_id:x}): {e}')
            return

        for signal in decoded_signals:
            x = timestamp
            y = decoded_signals[signal]
            if isinstance(y, NamedSignalValue):
                y = str(y)
            signal = message.name + '.' + signal
            self.signals.add_value(signal, x, y)

    def failed_to_parse_line(self, timestamp, line):
        if self.show_invalid_syntax:
            self.x_invalid_syntax.append(timestamp)
        if not self.ignore_invalid_syntax:
            print(f"Failed to parse line: {line!r}")

    # ------- at end -------

    def plot(self, xlabel):
        self.signals.plot(xlabel, self.x_invalid_syntax, self.x_unknown_frames, self.x_invalid_data)
        if self.output_filename:
            plt.savefig(self.output_filename)
            print(f"Result written to {self.output_filename}")
        else:
            plt.show()

class Signals:

    '''
    Parses the command line options which signals should be plotted
    and saves the corresponding values in Graph objects.
    Automatically inserts None values as specified by break_time.
    Plots the values using matplotlib.pyplot.
    '''

    # added between signal names used as default ylabel
    YLABEL_SEP = ', '

    # before re.escape
    SEP_SUBPLOT = '-'
    SEP_AXES = ','

    SEP_FMT = ':'
    FMT_STEM = '|'

    # after re.escape
    SEP_SG = re.escape('.')

    WILDCARD_MANY = re.escape('*')
    WILDCARD_ONE  = re.escape('?')

    COLOR_INVALID_SYNTAX = '#ff0000'
    COLOR_UNKNOWN_FRAMES = '#ffab00'
    COLOR_INVALID_DATA   = '#ff00ff'
    ERROR_LINEWIDTH = 1

    FIRST_SUBPLOT = 1
    FIRST_AXIS = 0

    # ------- initialization -------

    def __init__(self, signals, case_sensitive, break_time, global_subplot_args, auto_color_ylabels, dbase):
        self.dbase = dbase
        self.args = signals
        self.global_subplot_args = global_subplot_args
        self.signals = []
        self.values = {}
        self.re_flags = 0 if case_sensitive else re.IGNORECASE
        self.break_time = break_time
        self.break_time_uninit = True
        self.subplot = self.FIRST_SUBPLOT
        self.subplot_axis = self.FIRST_AXIS
        self.subplot_args = {}
        self.subplot_argparser = argparse.ArgumentParser()
        self.subplot_argparser.add_argument('signals', nargs='*')
        add_subplot_options(self.subplot_argparser)

        i0 = 0
        while True:
            try:
                i1 = signals.index(self.SEP_SUBPLOT, i0)
            except ValueError:
                i1 = None

            try:
                i12 = signals.index(self.SEP_AXES, i0)
            except ValueError:
                i12 = None
            if i1 is None or (i12 is not None and i12 < i1):
                i1 = i12

            subplot_signals = signals[i0:i1]
            subplot_args = self.subplot_argparser.parse_args(subplot_signals)
            if auto_color_ylabels and subplot_args.color is None:
                subplot_args.color = f"C{self.subplot_axis}"
            self.subplot_args[(self.subplot, self.subplot_axis)] = subplot_args
            self._ylabel = ""
            for sg in subplot_args.signals:
                self.add_signal(sg)
            if subplot_args.ylabel is None and self._ylabel:
                subplot_args.ylabel = self._ylabel

            if i1 is None:
                break

            if signals[i1] == self.SEP_SUBPLOT:
                self.subplot += 1
                self.subplot_axis = self.FIRST_AXIS
            else:
                self.subplot_axis += 1
            i0 = i1 + 1

        if not self.signals:
            self.add_signal('*')

        self.compile_reo()

    def init_break_time(self, datatype):
        if self.break_time <= 0:
            self.break_time = None
        elif datatype == datetime.datetime:
            self.half_break_time = datetime.timedelta(seconds=self.break_time/2)
            self.break_time = datetime.timedelta(seconds=self.break_time)
        else:
            self.half_break_time = self.break_time / 2
        self.break_time_uninit = False

    def add_signal(self, signal):
        if self.SEP_FMT in signal:
            signal, fmt = signal.split(self.SEP_FMT, 1)
            if fmt.startswith(self.FMT_STEM):
                fmt = fmt[len(self.FMT_STEM):]
                plt_func = 'stem'
            else:
                plt_func = 'plot'
        else:
            fmt = ''
            plt_func = 'plot'

        if self._ylabel:
            self._ylabel += self.YLABEL_SEP
        self._ylabel += signal

        signal = re.escape(signal)
        if self.SEP_SG not in signal:
            signal = self.WILDCARD_MANY + self.SEP_SG + signal
        signal = signal.replace(self.WILDCARD_MANY, '.*')
        signal = signal.replace(self.WILDCARD_ONE, '.')
        signal += '$'
        reo = re.compile(signal, self.re_flags)

        sgo = Signal(reo, self.subplot, self.subplot_axis, plt_func, fmt)
        self.signals.append(sgo)

    def compile_reo(self):
        self.reo = re.compile('|'.join(sg.reo.pattern for sg in self.signals), re.IGNORECASE)

    # ------- while reading data -------

    def add_value(self, signal, x, y):
        if not self.is_displayed_signal(signal):
            return

        if signal not in self.values:
            graph = Graph()
            self.values[signal] = graph
        else:
            graph = self.values[signal]
            last_x = graph.x[-1]
            if self.break_time_uninit:
                self.init_break_time(type(x))
            if self.break_time and last_x + self.break_time < x:
                x_break = last_x + self.half_break_time
                graph.x.append(x_break)
                graph.y.append(None)
        graph.x.append(x)
        graph.y.append(y)

    def is_displayed_signal(self, signal):
        return self.reo.match(signal)

    # ------- at end -------

    SUBPLOT_DIRECT_NAMES = ('title', 'ylabel')
    def plot(self, xlabel, x_invalid_syntax, x_unknown_frames, x_invalid_data):
        self.default_xlabel = xlabel
        splot = None
        last_subplot = self.FIRST_SUBPLOT - 1
        last_axis = None
        axis_format_uninitialized = True
        sorted_signal_names = sorted(self.values.keys())
        self.legend_handles = []
        self.legend_labels = []
        for sgo in self.signals:
            if sgo.subplot > last_subplot:
                if splot is None:
                    axes = None
                else:
                    axes = splot.axes
                    self.finish_subplot(splot, self.subplot_args[(last_subplot, last_axis)])

                splot = plt.subplot(self.subplot, 1, sgo.subplot, sharex=axes)

                last_subplot = sgo.subplot
                last_axis = sgo.axis
            elif sgo.axis > last_axis:
                self.finish_axis(splot, self.subplot_args[(last_subplot, last_axis)])
                splot = splot.twinx()
                last_axis = sgo.axis

            plotted = False
            for signal_name in sorted_signal_names:
                graph = self.values[signal_name]
                if not sgo.match(signal_name):
                    continue
                if graph.plotted_signal:
                    if not self.is_replotting_desired(sgo, graph.plotted_signal):
                        continue
                else:
                    graph.plotted_signal = sgo

                x = graph.x
                y = graph.y
                if axis_format_uninitialized and x:
                    if isinstance(x[0], float):
                        splot.axes.xaxis.set_major_formatter(lambda x,pos: str(datetime.timedelta(seconds=x)))
                    axis_format_uninitialized = False
                plt_func = getattr(splot, sgo.plt_func)
                if self.global_subplot_args.show_units and self.subplot_args[(sgo.subplot, sgo.axis)].show_units:
                    unit = self.get_signal_unit(signal_name)
                    signal_name = f"{signal_name} [{unit}]"
                container = plt_func(x, y, sgo.fmt, label=signal_name)
                color = self.subplot_args[(sgo.subplot, sgo.axis)].color
                if color is not None and self.contains_no_color(sgo.fmt):
                    for line in container:
                        line.set_color(color)
                plotted = True

            if not plotted:
                print(f"WARNING: signal {sgo.reo.pattern!r} with format {sgo.fmt!r} was not plotted.")

        self.plot_error(splot, x_invalid_syntax, 'invalid syntax', self.COLOR_INVALID_SYNTAX)
        self.plot_error(splot, x_unknown_frames, 'unknown frames', self.COLOR_UNKNOWN_FRAMES)
        self.plot_error(splot, x_invalid_data, 'invalid data', self.COLOR_INVALID_DATA)
        self.finish_subplot(splot, self.subplot_args[(last_subplot, last_axis)])

    def finish_axis(self, splot, subplot_args):
        kw = {key:val for key,val in vars(subplot_args).items() if val is not None and key in self.SUBPLOT_DIRECT_NAMES}
        for key in self.SUBPLOT_DIRECT_NAMES:
            if key not in kw:
                val = getattr(self.global_subplot_args, key)
                if val is not None:
                    kw[key] = val
        if kw:
            splot.set(**kw)

        if subplot_args.xlabel is not None:
            xlabel = subplot_args.xlabel
        elif self.global_subplot_args.xlabel is not None:
            xlabel = self.global_subplot_args.xlabel
        else:
            xlabel = self.default_xlabel
        splot.set_xlabel(xlabel)

        if subplot_args.ymin is None:
            subplot_args.ymin = self.global_subplot_args.ymin
        if subplot_args.ymax is None:
            subplot_args.ymax = self.global_subplot_args.ymax
        if subplot_args.ymin is not None or subplot_args.ymax is not None:
            splot.axes.set_ylim(subplot_args.ymin, subplot_args.ymax)

        if subplot_args.color is not None:
            splot.yaxis.label.set_color(subplot_args.color)
            splot.tick_params(axis='y', which='both', colors=subplot_args.color)

        handles, labels = splot.get_legend_handles_labels()
        self.legend_handles.extend(handles)
        self.legend_labels.extend(labels)

    def finish_subplot(self, splot, subplot_args):
        self.finish_axis(splot, subplot_args)
        splot.legend(self.legend_handles, self.legend_labels)
        self.legend_handles = []
        self.legend_labels = []

    def contains_no_color(self, fmt):
        for c in fmt:
            if c in PYPLOT_BASE_COLORS:
                return False
        return True

    def plot_error(self, splot, xs, label, color):
        if xs:
            label += f" ({len(xs)})"
            xs = iter(xs)
            splot.axvline(next(xs), color=color, linewidth=self.ERROR_LINEWIDTH, label=label)
            for x in xs:
                splot.axvline(x, color=color, linewidth=self.ERROR_LINEWIDTH)

    def is_replotting_desired(self, current_signal, previously_plotted_signal):
        if current_signal.reo.pattern == previously_plotted_signal.reo.pattern:
            # if the user bothers to type out the same regex twice
            # it is probably intended to be plotted twice
            return True
        # if the user bothers to type out a complete signal name without wildcards
        # he/she probably means to plot this signal even if it has been plotted already
        return '.' not in current_signal.reo.pattern

    def get_signal_unit(self, signal_name):
        msg, signal = re.split(self.SEP_SG, signal_name)
        return self.dbase.get_message_by_name(msg).get_signal_by_name(signal).unit

class Signal:

    '''
    Stores meta information about signals to be plotted:
    - a regex matching all signals it refers to
    - the format how it should be plotted
    - the subplot in which to display the signal

    It does *not* store the values to be plotted.
    They are stored in Graph.
    Signal and Graph have a one-to-many-relationship.
    '''

    # ------- initialization -------

    def __init__(
        self, reo: "re.Pattern[str]",
        subplot: int,
        axis: int,
        plt_func: str,
        fmt: str,
    ) -> None:
        self.reo = reo
        self.subplot = subplot
        self.axis = axis
        self.plt_func = plt_func
        self.fmt = fmt

    # ------- while reading data -------

    def match(self, signal):
        return self.reo.match(signal)

class Graph:

    '''
    A container for the values to be plotted.
    The corresponding signal names are the keys in Signals.values.
    The format how to plot this data is stored in Signals.signals (a list of Signal objects).

    plotted_signal stores a Signal object with which this graph has been plotted already
    to avoid undesired replotting of the same data in case the user gives two regex
    matching the same signal, one more specific to match a certain signal with a special format
    and one more generic matching the rest with another format.
    '''

    __slots__ = ('plotted_signal', 'x', 'y')

    def __init__(self):
        self.x = []
        self.y = []
        self.plotted_signal = None


class RawDescriptionArgumentDefaultsHelpFormatter(
    argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):
    pass


def add_subparser(subparsers):
    '''
    Is called from ../__init__.py.
    It adds the options for this subprogram to the argparse parser.
    It sets the entry point for this subprogram by setting a default values for func.
    '''
    plot_parser = subparsers.add_parser(
        'plot',
        description=__doc__,
        formatter_class=RawDescriptionArgumentDefaultsHelpFormatter)
    plot_parser.add_argument(
        '-c', '--no-decode-choices',
        action='store_true',
        help='Do not convert scaled values to choice strings.')
    plot_parser.add_argument(
        '-e', '--encoding',
        help='File encoding of dbc file.')
    plot_parser.add_argument(
        '-m', '--frame-id-mask',
        type=Integer(0),
        help=('Only compare selected frame id bits to find the message in the '
              'database. By default the candump and database frame ids must '
              'be equal for a match.'))
    plot_parser.add_argument(
        '-I', '--case-sensitive',
        action='store_true',
        help='Match the signal names case sensitive.')
    plot_parser.add_argument(
        '-l', '--line-numbers',
        action='store_true',
        help='Use line numbers instead of time stamps on the horizontal axis (useful with `candump -td`).')
    plot_parser.add_argument(
        '-t', '--break-time',
        default=100,
        type=float,
        help=('If the time distance between two consecutive signals is longer than this value '
              'the line in the plot will be interrupted. The value is given in seconds '
              '(if timestamps are used) or input lines (if line numbers are used). '
              '-1 means infinite. '))

    plot_parser.add_argument(
        '--show-invalid-syntax',
        action='store_true',
        help='Show a marker for lines which could not be parsed. This implies -l.')
    plot_parser.add_argument(
        '--show-unknown-frames',
        action='store_true',
        help='Show a marker for messages which are not contained in the database file.')
    plot_parser.add_argument(
        '--show-invalid-data',
        action='store_true',
        help='Show a marker for messages with data which could not be parsed.')
    plot_parser.add_argument(
        '-s', '--show-errors',
        action='store_true',
        help='Show all error messages in the plot. This is an abbreviation for all --show-* options. This implies -l.')

    plot_parser.add_argument(
        '--ignore-invalid-syntax',
        action='store_true',
        help='Don\'t print an error message for lines which could not be parsed.')
    plot_parser.add_argument(
        '--ignore-unknown-frames',
        action='store_true',
        help='Don\'t print an error message for messages which are not contained in the database file.')
    plot_parser.add_argument(
        '--ignore-invalid-data',
        action='store_true',
        help='Don\'t print an error message for messages with data which could not be parsed.')
    plot_parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='Don\'t print any error messages. This is an abbreviation for all --ignore-* options.')

    plot_parser.add_argument(
        '-o', '--output-file',
        help='A file to write the plot to instead of displaying it in a window.')

    plot_parser.add_argument(
        '-ss', '--start',
        help='A start time or line number. Everything before is ignored. '
             'This filters the lines/messages to be processed. It does *not* set the minimum value of the x-axis.')
    plot_parser.add_argument(
        '-to', '--stop',
        help='An end time or line number. Everything after is ignored. '
             'This filters the lines/messages to be processed. It does *not* set the maximum value of the x-axis.')

    plot_parser.add_argument(
        '--style',
        help='The matplotlib style to be used.')
    plot_parser.add_argument(
        '--list-styles',
        action='store_true',
        help='Print all available matplotlib styles without drawing a plot.')
    plot_parser.add_argument(
        '-ac', '--auto-color-ylabels',
        action='store_true',
        help='This is equivalent to applying --color C0 to the first y-axis, --color C1 to the second and so on.')
    plot_parser.add_argument(
        '--prune',
        action='store_true',
        help='Try to shorten the names of named signal choices.')
    plot_parser.add_argument(
        '--no-strict',
        action='store_true',
        help='Skip database consistency checks.')

    plot_parser.add_argument(
        'database',
        help='Database file.')
    plot_parser.add_argument(
        'signals',
        nargs='*',
        help='The signals to be plotted.')
    plot_parser.set_defaults(func=_do_decode)

    subplot_arggroup = plot_parser.add_argument_group('subplot arguments',
        '''\
The following options can be used to configure the subplots/axes.
If they shall apply to a specific subplot/axis they must be placed among the signals for that subplot/axis and a -- must mark the end of the global optional arguments.
Otherwise they are used as default value for each subplot/axis.
''')
    add_subplot_options(subplot_arggroup)

def add_subplot_options(arg_group):
    arg_group.add_argument('--title')
    arg_group.add_argument('--color',
        help='The color to be used for the y-label and the signals (unless a different color is given for the signal). '
             'All string formats explained in the following link are allowed: https://matplotlib.org/tutorials/colors/colors.html')
    arg_group.add_argument('--xlabel')
    arg_group.add_argument('--ylabel')
    arg_group.add_argument('--ymin', type=float)
    arg_group.add_argument('--ymax', type=float)
    arg_group.add_argument(
        '--no-units',
        dest='show_units',
        action='store_false',
        help='Do not show units in legend labels.')
    return arg_group



# Store the specific DBC format properties of objects

from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cantools.database.can.attribute import AttributeType
    from cantools.database.can.attribute_definition import (
        AttributeDefinitionType,
    )
    from cantools.database.can.environment_variable import EnvironmentVariable
    from cantools.typechecking import Choices

    from .dbc import DbcRelationAttributes

class DbcSpecifics:

    def __init__(self,
                 *,
                 attributes: OrderedDict[str, AttributeType] | None = None,
                 attribute_definitions: OrderedDict[str, AttributeDefinitionType] | None = None,
                 environment_variables: OrderedDict[str, EnvironmentVariable] | None = None,
                 value_tables: OrderedDict[str, Choices] | None = None,
                 relation_attributes: DbcRelationAttributes | None = None,
                 relation_attribute_definitions: OrderedDict[str, AttributeDefinitionType] | None = None) -> None:
        self._attributes = attributes or OrderedDict()
        self._attribute_definitions = attribute_definitions or OrderedDict()
        self._environment_variables = environment_variables or OrderedDict()
        self._value_tables = value_tables or OrderedDict()
        self._relation_attributes = relation_attributes
        self._relation_attribute_definitions = relation_attribute_definitions or OrderedDict()

    @property
    def attributes(self) -> OrderedDict[str, AttributeType]:
        """The DBC specific attributes of the parent object (database, node,
        message or signal) as a dictionary.

        """

        return self._attributes

    @property
    def attribute_definitions(self) -> OrderedDict[str, AttributeDefinitionType]:
        """The DBC specific attribute definitions as dictionary.

        """

        return self._attribute_definitions

    @property
    def value_tables(self) -> OrderedDict[str, Choices]:
        """An ordered dictionary of all value tables. Only valid for DBC
        specifiers on database level.

        """

        return self._value_tables

    @property
    def environment_variables(self) -> OrderedDict[str, EnvironmentVariable]:
        """An ordered dictionary of all environment variables. Only valid for
        DBC specifiers on database level.

        """

        return self._environment_variables

    @property
    def relation_attributes(self) -> DbcRelationAttributes | None:
        """The DBC specific relation attributes.

        """

        return self._relation_attributes

    @property
    def relation_attribute_definitions(self) -> OrderedDict[str, AttributeDefinitionType]:
        """The DBC specific attribute definitions rel as dictionary.

        """

        return self._relation_attribute_definitions





class NamedSignalValue:
    """Represents a named value of a signal.

    Named values map an integer number to a human-readable
    string. Some file formats like ARXML support specifying
    descriptions for the named value.
    """

    def __init__(
        self,
        value: int,
        name: str,
        comments: dict[str, str] | None = None,
    ) -> None:
        #: The text intended for human consumption which the specified integer
        #: is mapped to.
        self.name = name

        #: The integer value that gets mapped.
        self.value = value

        self._comments = comments or {}

    @property
    def comments(self) -> dict[str, str]:
        """The descriptions of the named value

        This is a dictionary containing the descriptions in multiple
        languages. The dictionary is indexed by the language.

        Example:

        .. code:: text

          # retrieve the English comment of the named value or an empty
          # string if none was specified.
          named_value.comments.get("EN", "")

        """

        return self._comments

    def __str__(self) -> str:
        return f"{self.name}"

    def __repr__(self) -> str:
        return f"'{self.name}'"

    def __eq__(self, x: object) -> bool:
        if isinstance(x, NamedSignalValue):
            return (
                x.value == self.value
                and x.name == self.name
                and x.comments == self.comments
            )
        elif isinstance(x, str):
            return x == self.name

        return False



# A DID.

import binascii
from typing import TypedDict

from cantools.database.diagnostics.data import Data
from cantools.typechecking import Formats, SignalDictType, SignalMappingType

from ..utils import create_encode_decode_formats, decode_data, encode_data


class _Codec(TypedDict):
    datas: list[Data]
    formats: Formats


class Did:
    """A DID with identifier and other information.

    """

    def __init__(self,
                 identifier: int,
                 name: str,
                 length: int,
                 datas: list[Data]) -> None:
        self._identifier = identifier
        self._name = name
        self._length = length
        self._datas = datas
        self.refresh()

    @property
    def identifier(self) -> int:
        """The did identifier as an integer.

        """

        return self._identifier

    @identifier.setter
    def identifier(self, value: int) -> None:
        self._identifier = value

    @property
    def name(self) -> str:
        """The did name as a string.

        """

        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def length(self) -> int:
        """The did length as an int.

        """

        return self._length

    @length.setter
    def length(self, value: int) -> None:
        self._length = value

    @property
    def datas(self) -> list[Data]:
        """The did datas as a string.

        """

        return self._datas

    @datas.setter
    def datas(self, value: list[Data]) -> None:
        self._datas = value

    def get_data_by_name(self, name: str) -> Data:
        for data in self._datas:
            if data.name == name:
                return data

        raise KeyError(name)

    def encode(self, data: SignalMappingType, scaling: bool = True) -> bytes:
        """Encode given data as a DID of this type.

        If `scaling` is ``False`` no scaling of datas is performed.

        >>> foo = db.get_did_by_name('Foo')
        >>> foo.encode({'Bar': 1, 'Fum': 5.0})
        b'\\x01\\x45\\x23\\x00\\x11'

        """

        encoded = encode_data(data,
                              self._codec['datas'],
                              self._codec['formats'],
                              scaling)
        encoded |= (0x80 << (8 * self._length))
        encoded_str = hex(encoded)[4:].rstrip('L')

        return binascii.unhexlify(encoded_str)[:self._length]

    def decode(self,
               data: bytes,
               decode_choices: bool = True,
               scaling: bool = True,
               allow_truncated: bool = False,
               allow_excess: bool = True) -> SignalDictType:
        """Decode given data as a DID of this type.

        If `decode_choices` is ``False`` scaled values are not
        converted to choice strings (if available).

        If `scaling` is ``False`` no scaling of datas is performed.

        >>> foo = db.get_did_by_name('Foo')
        >>> foo.decode(b'\\x01\\x45\\x23\\x00\\x11')
        {'Bar': 1, 'Fum': 5.0}

        """

        return decode_data(data[:self._length],
                           self.length,
                           self._codec['datas'],
                           self._codec['formats'],
                           decode_choices,
                           scaling,
                           allow_truncated,
                           allow_excess)

    def refresh(self) -> None:
        """Refresh the internal DID state.

        """

        self._codec = _Codec(
            datas=self._datas,
            formats=create_encode_decode_formats(self._datas,
                                                 self._length)
        )

    def __repr__(self) -> str:
        return f"did('{self._name}', 0x{self._identifier:04x})"



# The tester module.

import queue
import time
from collections import UserDict
from collections.abc import Mapping

import can

from cantools.database.can.message import Message as MessageCls
from cantools.database.can.message import SignalTreeType

from .errors import Error


class DecodedMessage:
    """A decoded message.

    """

    def __init__(self, name, signals):
        self.name = name
        self.signals = signals


class Messages(UserDict):
    def __setitem__(self, message_name, value):
        if getattr(self, '_frozen', False) and message_name not in self.data:
            raise KeyError(message_name)
        self.data[message_name] = value

    def __missing__(self, key):
        raise Error(f"invalid message name '{key}'")


def _invert_signal_tree(
        tree: SignalTreeType,
        cur_mpx: dict | None = None,
        ret: dict | None = None
) -> dict:

    """The tree is laid out with two kinds of dicts.  Single-element dict
    keyed by string -> multiplexer, which is own dict keyed by
    integers.

    """

    if ret is None:
        ret = {}

    if cur_mpx is None:
        cur_mpx = {}

    for sigs in tree:
        if isinstance(sigs, Mapping): # outer signal keyed by muliplexer name
            (mpx_name, mpx_vals), = sigs.items()
            for mpx_val, sig_tree in mpx_vals.items(): # inner signal
                #                          keyed by multiplexer values
                next_mpx = cur_mpx.copy()
                next_mpx[mpx_name] = mpx_val
                _invert_signal_tree(sig_tree, next_mpx, ret)

        elif isinstance(sigs, str):
            ret.setdefault(sigs,[]).append(set(cur_mpx.items()))
        else:
            raise TypeError(repr(sigs))

    return ret

def invert_signal_tree(tree: SignalTreeType) -> dict:
    """Return a mapping of signals to the multiplex settings that will
    yield the signal.

    {signal: [{mplexers}, {mplexers...}]}

    """
    return _invert_signal_tree(tree)

class Listener(can.Listener):

    def __init__(self, database, messages, input_queue, on_message):
        self._database = database
        self._messages = messages
        self._input_queue = input_queue
        self._on_message = on_message

    def on_message_received(self, msg):
        if msg.is_error_frame or msg.is_remote_frame:
            return

        try:
            database_message = self._database.get_message_by_frame_id(msg.arbitration_id, msg.is_extended_id)
        except KeyError:
            return

        if database_message.name not in self._messages:
            return

        message = self._messages[database_message.name]

        if not message.enabled:
            return

        decoded = DecodedMessage(database_message.name,
                                 database_message.decode(msg.data,
                                                         message.decode_choices,
                                                         message.scaling))

        if self._on_message:
            self._on_message(decoded)

        self._input_queue.put(decoded)


class Message(UserDict):

    def __init__(self,
                 database: MessageCls,
                 can_bus: can.BusABC,
                 input_list: list[MessageCls],
                 input_queue: queue.Queue[DecodedMessage],
                 decode_choices: bool,
                 scaling: bool,
                 padding: bool,
                 strict: bool = True) -> None:
        super().__init__()
        self.database = database
        self._mplex_map = invert_signal_tree(database.signal_tree)
        self._can_bus = can_bus
        self._input_queue = input_queue
        self.decode_choices = decode_choices
        self.scaling = scaling
        self.padding = padding
        self.strict = strict
        self._input_list = input_list
        self.enabled = True
        self._can_message = None
        self._periodic_task = None
        self._signal_names = {s.name for s in self.database.signals}
        self.update(self._prepare_initial_signal_values())

    @property
    def periodic(self):
        return self.database.cycle_time is not None

    def __getitem__(self, signal_name):
        return self.data[signal_name]

    def __setitem__(self, signal_name, value):
        if signal_name not in self._signal_names:
            raise KeyError(signal_name)
        self.data[signal_name] = value
        self._update_can_message()

    def update(self, signals):
        s = dict(signals)
        new_signal_names = set(s) - self._signal_names
        if new_signal_names:
            raise KeyError(repr(new_signal_names))

        self.data.update(s)
        self._update_can_message()

    def send(self, signals=None):
        if signals is not None:
            self.update(signals)

        self._can_bus.send(self._can_message)

    def expect(self, signals=None, timeout=None, discard_other_messages=True):
        if signals is None:
            signals = {}

        decoded = self._expect_input_list(signals, discard_other_messages)

        if decoded is None:
            decoded = self._expect_input_queue(signals,
                                               timeout,
                                               discard_other_messages)

        return decoded

    def _expect_input_list(self, signals, discard_other_messages):
        other_messages = []

        while len(self._input_list) > 0:
            message = self._input_list.pop(0)
            decoded = self._filter_expected_message(message, signals)

            if decoded is not None:
                break

            other_messages.append(message)
        else:
            decoded = None

        if not discard_other_messages:
            other_messages += self._input_list
            del self._input_list[:]
            self._input_list.extend(other_messages)

        return decoded

    def _expect_input_queue(self, signals, timeout, discard_other_messages):
        if timeout is not None:
            end_time = time.time() + timeout
            remaining_time = timeout
        else:
            remaining_time = None

        while True:
            try:
                message = self._input_queue.get(timeout=remaining_time)
            except queue.Empty:
                return

            decoded = self._filter_expected_message(message, signals)

            if decoded is not None:
                return decoded

            if not discard_other_messages:
                self._input_list.append(message)

            if timeout is not None:
                remaining_time = end_time - time.time()

                if remaining_time <= 0:
                    return

    def _filter_expected_message(self, message, signals):
        if message.name == self.database.name and all(message.signals[name] == signals[name] for name in signals):
            return message.signals

    def send_periodic_start(self):
        if not self.enabled:
            return

        self._periodic_task = self._can_bus.send_periodic(
            self._can_message,
            self.database.cycle_time / 1000.0)

    def send_periodic_stop(self):
        if self._periodic_task is not None:
            self._periodic_task.stop()
            self._periodic_task = None

    def _update_can_message(self):
        arbitration_id = self.database.frame_id
        extended_id = self.database.is_extended_frame
        pruned_data = self.database.gather_signals(self.data)
        data = self.database.encode(pruned_data,
                                    self.scaling,
                                    self.padding,
                                    strict=self.strict)
        self._can_message = can.Message(arbitration_id=arbitration_id,
                                        is_extended_id=extended_id,
                                        data=data,
                                        is_fd=self.database.is_fd,
                                        dlc=self.database.length,
                                        check=True)

        if self._periodic_task is not None:
            self._periodic_task.modify_data(self._can_message)

    def _prepare_initial_signal_values(self):
        initial_sig_values = {}

        # Choose a valid set of mux settings
        mplex_settings = {}
        for m0 in reversed(self._mplex_map.values()):
            for m1 in m0:
                mplex_settings.update(m1)

        for signal in self.database.signals:
            minimum = 0 if not signal.minimum else signal.minimum
            maximum = 0 if not signal.maximum else signal.maximum
            if signal.initial:
                # use initial signal value (if set)
                initial_sig_values[signal.name] = signal.initial
            elif signal.is_multiplexer:
                initial_sig_values[signal.name] = mplex_settings.get(signal.name, 0)
            elif minimum <= 0 <= maximum:
                # use 0 if in allowed range
                initial_sig_values[signal.name] = 0
            else:
                # set at least some default value
                initial_sig_values[signal.name] = minimum
        return initial_sig_values


class Tester:
    """Test given node `dut_name` on given CAN bus `bus_name`.

    `database` is a :class:`~cantools.database.can.Database` instance.

    `can_bus` a CAN bus object, normally created using the python-can
    package.

    The `on_message` callback is called for every successfully decoded
    received message. It is called with one argument, an
    :class:`~cantools.tester.DecodedMessage` instance.

    Here is an example of how to create a tester:

    >>> import can
    >>> import cantools
    >>> can.rc['interface'] = 'socketcan'
    >>> can.rc['channel'] = 'vcan0'
    >>> can_bus = can.interface.Bus()
    >>> database = cantools.database.load_file('tests/files/tester.kcd')
    >>> tester = cantools.tester.Tester('PeriodicConsumer', database, can_bus, 'PeriodicBus')

    """

    def __init__(self,
                 dut_name,
                 database,
                 can_bus,
                 bus_name=None,
                 on_message=None,
                 decode_choices=True,
                 scaling=True,
                 padding=False,
                 strict=True):
        self._dut_name = dut_name
        self._bus_name = bus_name
        self._database = database
        self._can_bus = can_bus
        self._input_list = []
        self._input_queue = queue.Queue()
        self._messages = Messages()
        self._is_running = False

        # DUT name validation.
        node_names = [node.name for node in database.nodes]

        if dut_name and not any(name == dut_name for name in node_names):
            raise Error(f"expected DUT name in {node_names}, but got '{dut_name}'")

        # BUS name validation.
        bus_names = [bus.name for bus in database.buses]

        if len(bus_names) == 0:
            if bus_name is not None:
                raise Error(
                    f"expected bus name None as there are no buses defined in "
                    f"the database, but got '{bus_name}'")
        elif not any(name == bus_name for name in bus_names):
            raise Error(f"expected bus name in {bus_names}, but got '{bus_name}'")

        for message in database.messages:
            if message.bus_name == bus_name:
                self._messages[message.name] = Message(message,
                                                       can_bus,
                                                       self._input_list,
                                                       self._input_queue,
                                                       decode_choices,
                                                       scaling,
                                                       padding,
                                                       strict=strict)

        listener = Listener(self._database,
                            self._messages,
                            self._input_queue,
                            on_message)
        self._notifier = can.Notifier(can_bus, [listener])
        self._messages._frozen = True

    def start(self):
        """Start the tester. Starts sending enabled periodic messages.

        >>> tester.start()

        """

        for message in self._messages.values():
            if self._dut_name and self._dut_name in message.database.senders:
                continue

            if not message.periodic:
                continue

            message.send_periodic_start()

        self._is_running = True

    def stop(self):
        """Stop the tester. Periodic messages will not be sent after this
        call. Call :meth:`~cantools.tester.Tester.start()` to resume a
        stopped tester.

        >>> tester.stop()

        """

        for message in self._messages.values():
            message.send_periodic_stop()

        self._is_running = False

    @property
    def messages(self):
        """Set and get signals in messages. Set signals takes effect
        immediately for started enabled periodic messages. Call
        :meth:`~cantools.tester.Tester.send()` for other messages.

        >>> periodic_message = tester.messages['PeriodicMessage1']
        >>> periodic_message
        {'Signal1': 0, 'Signal2': 0}
        >>> periodic_message['Signal1'] = 1
        >>> periodic_message.update({'Signal1': 2, 'Signal2': 5})
        >>> periodic_message
        {'Signal1': 2, 'Signal2': 5}

        """

        return self._messages

    def enable(self, message_name):
        """Enable given message `message_name` and start sending it if its
        periodic and the tester is running.

        >>> tester.enable('PeriodicMessage1')

        """

        message = self._messages[message_name]
        message.enabled = True

        if self._is_running and message.periodic:
            message.send_periodic_start()

    def disable(self, message_name):
        """Disable given message `message_name` and stop sending it if its
        periodic, enabled and the tester is running.

        >>> tester.disable('PeriodicMessage1')

        """

        message = self._messages[message_name]
        message.enabled = False

        if self._is_running and message.periodic:
            message.send_periodic_stop()

    def send(self, message_name, signals=None):
        """Send given message `message_name` and optional signals `signals`.

        >>> tester.send('Message1', {'Signal2': 10})
        >>> tester.send('Message1')

        """

        self._messages[message_name].send(signals)

    def expect(self,
               message_name,
               signals=None,
               timeout=None,
               discard_other_messages=True):
        """Expect given message `message_name` and signal values `signals`
        within `timeout` seconds.

        Give `signals` as ``None`` to expect any signal values.

        Give `timeout` as ``None`` to wait forever.

        Messages are read from the input queue, and those not matching
        given `message_name` and `signals` are discarded if
        `discard_other_messages` is
        ``True``. :meth:`~cantools.tester.Tester.flush_input()` may be
        called to discard all old messages in the input queue before
        calling the expect function.

        Returns the expected message, or ``None`` on timeout.

        >>> tester.expect('Message2', {'Signal1': 13})
        {'Signal1': 13, 'Signal2': 9}

        """

        return self._messages[message_name].expect(signals,
                                                   timeout,
                                                   discard_other_messages)

    def flush_input(self):
        """Flush, or discard, all messages in the input queue.

        """

        del self._input_list[:]

        while not self._input_queue.empty():
            self._input_queue.get()



# Utility functions.

import os.path
import re
from collections import OrderedDict
from collections.abc import Callable, Sequence
from typing import (
    TYPE_CHECKING,
    Final,
    Literal,
    Union,
)

from ..typechecking import (
    ByteOrder,
    Choices,
    Formats,
    SignalDictType,
    SignalMappingType,
    SignalValueType,
)
from .errors import DecodeError, EncodeError

if TYPE_CHECKING:
    from ..database import Database
    from ..database.can.attribute import AttributeType
    from ..database.can.environment_variable import EnvironmentVariable
    from ..database.can.message import Message
    from ..database.can.node import Node
    from ..database.can.signal import Signal
    from ..database.diagnostics import Data

try:
    import bitstruct.c  # type: ignore
except ImportError:
    import bitstruct  # type: ignore


def format_or(items: list[int | str]) -> str:
    string_items = [str(item) for item in items]

    if len(string_items) == 1:
        return string_items[0]
    else:
        return '{} or {}'.format(', '.join(string_items[:-1]),
                                 string_items[-1])


def format_and(items: Sequence[int | str]) -> str:
    string_items = [str(item) for item in items]

    if len(string_items) == 1:
        return str(string_items[0])
    else:
        return '{} and {}'.format(', '.join(string_items[:-1]),
                                  string_items[-1])


def start_bit(signal: Union["Signal", "Data"]) -> int:
    if signal.byte_order == 'big_endian':
        return 8 * (signal.start // 8) + (7 - (signal.start % 8))
    else:
        return signal.start


def _encode_signal_values(signals: Sequence[Union["Signal", "Data"]],
                          signal_values: SignalMappingType,
                          scaling: bool,
                          ) -> dict[str, int | float]:
    """
    Convert a dictionary of physical signal values into raw ones.
    """
    raw_values: dict[str, int | float] = {}
    for signal in signals:
        name = signal.name
        conversion = signal.conversion
        value = signal_values[name]

        if isinstance(value, (int, float)):
            if scaling:
                raw_values[name] = conversion.numeric_scaled_to_raw(value)
                continue

            raw_values[name] = value if conversion.is_float else round(value)
            continue
        elif isinstance(value, str):
            raw_values[name] = conversion.choice_to_number(value)
            continue
        else:
            # validate the given NamedSignalValue first
            if value != conversion.raw_to_scaled(value.value, decode_choices=True):
                raise EncodeError(
                    f"Invalid 'NamedSignalValue' name/value pair not found! Name {value.name}, value {value.value}"
                )

            raw_values[name] = value.value
            continue

        raise EncodeError(
            f"Unable to encode signal '{name}' "
            f"with type '{value.__class__.__name__}'."
        )

    return raw_values


def encode_data(signal_values: SignalMappingType,
                signals: Sequence[Union["Signal", "Data"]],
                formats: Formats,
                scaling: bool
                ) -> int:
    if len(signals) == 0:
        return 0

    raw_signal_values = _encode_signal_values(signals, signal_values, scaling)
    big_packed = formats.big_endian.pack(raw_signal_values)
    little_packed = formats.little_endian.pack(raw_signal_values)
    packed_union = int.from_bytes(big_packed, "big") | int.from_bytes(little_packed, "little")

    return packed_union


def decode_data(data: bytes,
                expected_length: int,
                signals: Sequence[Union["Signal", "Data"]],
                formats: Formats,
                decode_choices: bool,
                scaling: bool,
                allow_truncated: bool,
                allow_excess: bool,
                ) -> SignalDictType:

    actual_length = len(data)
    if actual_length != expected_length:
        if allow_truncated:
            # pad the data with 0xff to prevent the codec from
            # raising an exception. Note that all signals
            # that contain garbage will be removed below.
            data = data.ljust(expected_length, b"\xFF")

        if allow_excess:
            # trim the payload data to match the expected size
            data = data[:expected_length]

        if len(data) != expected_length:
            raise DecodeError(f"Wrong data size: {actual_length} instead of "
                              f"{expected_length} bytes")

    try:
        unpacked = {
            **formats.big_endian.unpack(data),
            **formats.little_endian.unpack(data[::-1]),
        }
    except (bitstruct.Error, ValueError) as e:
        # bitstruct returns different errors in PyPy and cpython
        raise DecodeError("unpacking failed") from e

    if actual_length < expected_length and allow_truncated:
        # remove signals that are outside available data bytes
        actual_bit_count = actual_length * 8
        for signal in signals:
            if signal.byte_order == "little_endian":
                sequential_start_bit = signal.start
            else:
                # Calculate start bit with inverted indices.
                # Function body of ``sawtooth_to_network_bitnum()``
                # is inlined for improved performance.
                sequential_start_bit = (8 * (signal.start // 8)) + (7 - (signal.start % 8))

            if sequential_start_bit + signal.length > actual_bit_count:
                del unpacked[signal.name]

    # scale the signal values and decode choices
    decoded: dict[str, SignalValueType] = {}
    for signal in signals:
        if (value := unpacked.get(signal.name)) is None:
            # signal value was removed above...
            continue

        if scaling:
            decoded[signal.name] = signal.conversion.raw_to_scaled(value, decode_choices)
        elif (decode_choices
              and signal.conversion.choices
              and (choice := signal.conversion.choices.get(value, None)) is not None):
            decoded[signal.name] = choice
        else:
            decoded[signal.name] = value

    return decoded


def create_encode_decode_formats(signals: Sequence[Union["Data", "Signal"]], number_of_bytes: int) -> Formats:
    format_length = (8 * number_of_bytes)

    def get_format_string_type(signal: Union["Data", "Signal"]) -> str:
        if signal.conversion.is_float:
            return 'f'
        elif signal.is_signed:
            return 's'
        else:
            return 'u'

    def padding_item(length: int) -> tuple[str, str, None]:
        fmt = f'p{length}'
        padding_mask = '1' * length

        return fmt, padding_mask, None

    def data_item(signal: Union["Data", "Signal"]) -> tuple[str, str, str]:
        fmt = f'{get_format_string_type(signal)}{signal.length}'
        padding_mask = '0' * signal.length

        return fmt, padding_mask, signal.name

    def fmt(items: list[tuple[str, str, str | None]]) -> str:
        return ''.join([item[0] for item in items])

    def names(items:  list[tuple[str, str, str | None]]) -> list[str]:
        return [item[2] for item in items if item[2] is not None]

    def padding_mask(items: list[tuple[str, str, str | None]]) -> int:
        try:
            return int(''.join([item[1] for item in items]), 2)
        except ValueError:
            return 0

    def create_big() -> tuple[str, int, list[str]]:
        items: list[tuple[str, str, str | None]] = []
        start = 0

        # Select BE signals
        be_signals = [signal for signal in signals if signal.byte_order == "big_endian"]

        # Ensure BE signals are sorted in network order
        sorted_signals = sorted(be_signals, key = lambda signal: sawtooth_to_network_bitnum(signal.start))

        for signal in sorted_signals:

            padding_length = (start_bit(signal) - start)

            if padding_length > 0:
                items.append(padding_item(padding_length))

            items.append(data_item(signal))
            start = (start_bit(signal) + signal.length)

        if start < format_length:
            length = format_length - start
            items.append(padding_item(length))

        return fmt(items), padding_mask(items), names(items)

    def create_little() -> tuple[str, int, list[str]]:
        items: list[tuple[str, str, str | None]] = []
        end = format_length

        for signal in signals[::-1]:
            if signal.byte_order == 'big_endian':
                continue

            padding_length = end - (signal.start + signal.length)

            if padding_length > 0:
                items.append(padding_item(padding_length))

            items.append(data_item(signal))
            end = signal.start

        if end > 0:
            items.append(padding_item(end))

        value = padding_mask(items)

        if format_length > 0:
            length = len(''.join([item[1] for item in items]))
            _packed = bitstruct.pack(f'u{length}', value)
            value = int.from_bytes(_packed, "little")

        return fmt(items), value, names(items)

    big_fmt, big_padding_mask, big_names = create_big()
    little_fmt, little_padding_mask, little_names = create_little()

    try:
        big_compiled = bitstruct.c.compile(big_fmt, big_names)
    except Exception:
        big_compiled = bitstruct.compile(big_fmt, big_names)

    try:
        little_compiled = bitstruct.c.compile(little_fmt, little_names)
    except Exception:
        little_compiled = bitstruct.compile(little_fmt, little_names)

    return Formats(big_compiled,
                   little_compiled,
                   big_padding_mask & little_padding_mask)


def sawtooth_to_network_bitnum(sawtooth_bitnum: int) -> int:
    '''Convert SawTooth bit number to Network bit number

    Byte     |   0   |   1   |
    Sawtooth |7 ... 0|15... 8|
    Network  |0 ... 7|8 ...15|
    '''
    return (8 * (sawtooth_bitnum // 8)) + (7 - (sawtooth_bitnum % 8))


def cdd_offset_to_dbc_start_bit(cdd_offset: int, bit_length: int, byte_order: ByteOrder) -> int:
    '''Convert CDD/c-style field bit offset to DBC field start bit convention.

    BigEndian (BE) fields are located by their MSBit's sawtooth index.
    LitteleEndian (LE) fields located by their LSBit's sawtooth index.
    '''
    if byte_order == "big_endian":
        # Note: Allow for BE fields that are smaller or larger than 8 bits.
        return (8 * (cdd_offset // 8)) + min(7, (cdd_offset % 8) + bit_length - 1)
    else:
        return cdd_offset


def prune_signal_choices(signal: "Signal") -> None:
    '''Shorten the names of the signal choices of a single signal

    For signals with multiple named values this means removing the
    longest common prefix that ends with an underscore and for which
    the removal still result the named signal values to be valid
    python identifiers. For signals with a single named choice, this
    means removing all leading segments between underscores which
    occur before a segment that contains a digit.

    Examples:

    ..code:: text

       MyMessage_MySignal_Uint32_Choice1, MyMessage_MySignal_Uint32_Choice2
       -> Choice1, Choice2
       MyMessage_MySignal_Uint32_NotAvailable
       -> NotAvailable

    '''

    if signal.choices is None:
        # no named choices
        return

    if len(signal.choices) == 1:
        # signal exhibits only a single named value: Use the longest
        # postfix starting with an underscore that does not contain
        # digits as the new name. If no such suffix exists, leave the
        # choice alone...
        key = next(iter(signal.choices.keys()))
        choice = next(iter(signal.choices.values()))
        m = re.match(r'^[0-9A-Za-z_]*?_([A-Za-z_]+)$', str(choice))
        val = str(choice)
        if m:
            val = m.group(1)

        if isinstance(choice, str):
            signal.choices[key] = val
        else:
            # assert isinstance(choice, NamedSignalValue)
            choice.name = val
        return

    # if there are multiple choices, remove the longest common prefix
    # that ends with an underscore from all of them provided that the
    # names of the choices stay valid identifiers
    choice_values = [ str(x) for x in signal.choices.values() ]
    full_prefix = os.path.commonprefix(choice_values)
    i = full_prefix.rfind('_')

    if i >= 0:
        full_prefix = full_prefix[0:i]
    else:
        # full_prefix does not contain an underscore
        # but the following algorithm assumes it does
        # and would strip too much
        return

    if not full_prefix:
        # the longest possible prefix is empty, i.e., there is nothing
        # to strip from the names of the signal choices
        return

    full_prefix_segments = full_prefix.split('_')

    # find the longest prefix of the choices which keeps all
    # names valid python identifiers
    prefix = ''
    n = 0
    valid_name_re = re.compile('^[a-zA-Z_][a-zA-Z0-9_]*$')
    for i in range(len(full_prefix_segments), -1, -1):
        if i == 0:
            # there is no such non-empty prefix
            return

        prefix = '_'.join(full_prefix_segments[:i]) + '_'
        n = len(prefix)

        if all(valid_name_re.match(x[n:]) for x in choice_values):
            break

    # remove the prefix from the choice names
    for key, choice in signal.choices.items():
        if isinstance(choice, str):
            signal.choices[key] = choice[n:]
        else:
            # assert isinstance(choice, NamedSignalValue)
            choice.name = choice.name[n:]


def prune_database_choices(database: "Database") -> None:
    '''
    Prune names of all named signal values of all signals of a database
    '''
    for message in database.messages:

        for signal in message.signals:
            prune_signal_choices(signal)

        if message.is_container:
            for cm in message.contained_messages:
                for cs in cm.signals:
                    prune_signal_choices(cs)


SORT_SIGNALS_DEFAULT: Final = 'default'
type_sort_signals = Callable[[list["Signal"]], list["Signal"]] | Literal['default'] | None

type_sort_attribute = \
    tuple[Literal['dbc'],     "AttributeType", None,   None,      None,     None] | \
    tuple[Literal['node'],    "AttributeType", "Node", None,      None,     None] | \
    tuple[Literal['message'], "AttributeType", None,   "Message", None,     None] | \
    tuple[Literal['signal'],  "AttributeType", None,   "Message", "Signal", None] | \
    tuple[Literal['envvar'],  "AttributeType", None,   None,      None,     "EnvironmentVariable"]

type_sort_attributes = Callable[[list[type_sort_attribute]], list[type_sort_attribute]] | Literal['default'] | None

type_sort_choices = Callable[[Choices], Choices] | None

def sort_signals_by_start_bit(signals: list["Signal"]) -> list["Signal"]:
    return sorted(signals, key=start_bit)


def sort_signals_by_start_bit_reversed(signals: list["Signal"]) -> list["Signal"]:
    return sorted(signals, key=start_bit)[::-1]


def sort_signals_by_name(signals: list["Signal"]) -> list["Signal"]:
    return sorted(signals, key=lambda s: s.name)


def sort_signals_by_start_bit_and_mux(signals: list["Signal"]) -> list["Signal"]:
    # sort by start bit
    signals = sorted(signals, key=start_bit)
    # but unmuxed values come first
    signals = sorted(signals, key=lambda s: bool(s.multiplexer_ids))
    # and group by mux... -1 is fine as the "no mux" case because even negative
    # multiplexors get cast to unsigned in the .dbc
    signals = sorted(
        signals, key=lambda s: s.multiplexer_ids[0] if s.multiplexer_ids else -1
    )

    return signals


def sort_choices_by_value(choices: Choices) -> Choices:
    return OrderedDict(sorted(choices.items(), key=lambda x: x[0]))


def sort_choices_by_value_descending(choices: Choices) -> Choices:
    return OrderedDict(sorted(choices.items(), key=lambda x: x[0], reverse=True))



# A CAN message.

import logging
from collections.abc import MutableSequence, Sequence
from copy import deepcopy
from typing import (
    Optional,
    cast,
)

from cantools.database.can.formats.dbc_specifics import DbcSpecifics

from ...typechecking import (
    Codec,
    Comments,
    ContainerDecodeResultListType,
    ContainerDecodeResultType,
    ContainerEncodeInputType,
    ContainerHeaderSpecType,
    ContainerUnpackListType,
    ContainerUnpackResultType,
    DecodeResultType,
    EncodeInputType,
    SignalDictType,
    SignalMappingType,
    SignalValueType,
)
from ..errors import DecodeError, EncodeError, Error
from ..namedsignalvalue import NamedSignalValue
from ..utils import (
    SORT_SIGNALS_DEFAULT,
    create_encode_decode_formats,
    decode_data,
    encode_data,
    format_or,
    sort_signals_by_start_bit,
    start_bit,
    type_sort_signals,
)
from .formats.arxml.message_specifics import AutosarMessageSpecifics
from .signal import Signal
from .signal_group import SignalGroup

LOGGER = logging.getLogger(__name__)

SignalTreeMuxElemType = dict[str, dict[int, Sequence[str]]]
SignalTreeElemType = str | SignalTreeMuxElemType
SignalTreeType = Sequence[SignalTreeElemType]


class Message:
    """A CAN message with frame id, comment, signals and other
    information.

    If `strict` is ``True`` an exception is raised if any signals are
    overlapping or if they don't fit in the message.

    By default signals are sorted by their start bit when their Message object is created.
    If you don't want them to be sorted pass `sort_signals = None`.
    If you want the signals to be sorted in another way pass something like
    `sort_signals = lambda signals: list(sorted(signals, key=lambda sig: sig.name))`
    """

    def __init__(self,
                 frame_id: int,
                 name: str,
                 length: int,
                 signals: list[Signal],
                 # if the message is a container message, this lists
                 # the messages which it potentially features
                 contained_messages: list['Message'] | None = None,
                 # header ID of message if it is part of a container message
                 header_id: int | None = None,
                 header_byte_order: str = 'big_endian',
                 unused_bit_pattern: int = 0x00,
                 comment: str | Comments | None = None,
                 senders: list[str] | None = None,
                 send_type: str | None = None,
                 cycle_time: int | None = None,
                 dbc_specifics: DbcSpecifics | None = None,
                 autosar_specifics: AutosarMessageSpecifics | None = None,
                 is_extended_frame: bool = False,
                 is_fd: bool = False,
                 bus_name: str | None = None,
                 signal_groups: list[SignalGroup] | None = None,
                 strict: bool = True,
                 protocol: str | None = None,
                 sort_signals: type_sort_signals = sort_signals_by_start_bit,
                 ) -> None:
        frame_id_bit_length = frame_id.bit_length()

        if is_extended_frame:
            if frame_id_bit_length > 29:
                raise Error(
                    f'Extended frame id 0x{frame_id:x} is more than 29 bits in '
                    f'message {name}.')
        elif frame_id_bit_length > 11:
            raise Error(
                f'Standard frame id 0x{frame_id:x} is more than 11 bits in '
                f'message {name}.')

        self._frame_id = frame_id
        self._header_id = header_id
        self._header_byte_order = header_byte_order
        self._is_extended_frame = is_extended_frame
        self._is_fd = is_fd
        self._name = name
        self._length = length
        self._unused_bit_pattern = unused_bit_pattern
        if sort_signals == SORT_SIGNALS_DEFAULT:
            self._signals = sort_signals_by_start_bit(signals)
        elif callable(sort_signals):
            self._signals = sort_signals(signals)
        else:
            self._signals = signals
        self._signal_dict: dict[str, Signal] = {}
        self._contained_messages = contained_messages or []

        # if the 'comment' argument is a string, we assume that is an
        # english comment. this is slightly hacky because the
        # function's behavior depends on the type of the passed
        # argument, but it is quite convenient...
        self._comments: Comments | None
        if isinstance(comment, str):
            # use the first comment in the dictionary as "The" comment
            self._comments = Comments({None: comment})
        else:
            # assume that we have either no comment at all or a
            # multi-lingual dictionary
            self._comments = comment

        self._senders = senders or []
        self._send_type = send_type
        self._cycle_time = cycle_time
        self._dbc = dbc_specifics
        self._autosar = autosar_specifics
        self._bus_name = bus_name
        self._signal_groups = signal_groups
        self._codecs: Codec | None = None
        self._signal_tree: SignalTreeType = []
        self._strict = strict
        self._protocol = protocol
        self.refresh()

    def _create_codec(self,
                      parent_signal: str | None = None,
                      multiplexer_id: int | None = None,
                      ) -> Codec:
        """Create a codec of all signals with given parent signal. This is a
        recursive function.

        """

        signals: list[Signal] = []
        multiplexers: dict[str, dict[int, Codec]] = {}

        # Find all signals matching given parent signal name and given
        # multiplexer id. Root signals' parent and multiplexer id are
        # both None.
        for signal in self._signals:
            if signal.multiplexer_signal != parent_signal:
                continue

            if (
                    multiplexer_id is not None
                    and (signal.multiplexer_ids is None or multiplexer_id not in signal.multiplexer_ids)
            ):
                continue

            if signal.is_multiplexer:
                children_ids: set[int] = set()

                for s in self._signals:
                    if s.multiplexer_signal != signal.name:
                        continue

                    if s.multiplexer_ids is not None:
                        children_ids.update(s.multiplexer_ids)

                # Some CAN messages will have muxes containing only
                # the multiplexer and no additional signals. At Tesla
                # these are indicated in advance by assigning them an
                # enumeration. Here we ensure that any named
                # multiplexer is included, even if it has no child
                # signals.
                if signal.conversion.choices:
                    children_ids.update(signal.conversion.choices.keys())

                for child_id in children_ids:
                    codec = self._create_codec(signal.name, child_id)

                    if signal.name not in multiplexers:
                        multiplexers[signal.name] = {}

                    multiplexers[signal.name][child_id] = codec

            signals.append(signal)

        return {
            'signals': signals,
            'formats': create_encode_decode_formats(signals,
                                                    self._length),
            'multiplexers': multiplexers
        }

    def _create_signal_tree(self, codec: Codec) -> SignalTreeType:
        """Create a multiplexing tree node of given codec. This is a recursive
        function.

        """

        nodes: list[SignalTreeElemType] = []

        for signal in codec['signals']:
            multiplexers = codec['multiplexers']

            node: str | dict[str, dict[int, SignalTreeType]]
            if signal.name in multiplexers:
                node = {
                    signal.name: {
                        mux: self._create_signal_tree(mux_codec)
                        for mux, mux_codec in multiplexers[signal.name].items()
                    }
                }
            else:
                node = signal.name

            nodes.append(cast('SignalTreeElemType', node))

        return nodes

    @property
    def header_id(self) -> int | None:
        """The header ID of the message if it is part of a container message.

        """

        return self._header_id

    @header_id.setter
    def header_id(self, value: int) -> None:
        self._header_id = value

    @property
    def header_byte_order(self) -> str:
        """The byte order of the header ID of the message if it is part of a
        container message.

        """

        return self._header_byte_order

    @header_byte_order.setter
    def header_byte_order(self, value: str) -> None:
        self._header_byte_order = value

    @property
    def frame_id(self) -> int:
        """The message frame id.

        """

        return self._frame_id

    @frame_id.setter
    def frame_id(self, value: int) -> None:
        self._frame_id = value

    @property
    def is_extended_frame(self) -> bool:
        """``True`` if the message is an extended frame, ``False`` otherwise.

        """

        return self._is_extended_frame

    @is_extended_frame.setter
    def is_extended_frame(self, value: bool) -> None:
        self._is_extended_frame = value

    @property
    def is_fd(self) -> bool:
        """``True`` if the message requires CAN-FD, ``False`` otherwise.

        """

        return self._is_fd

    @is_fd.setter
    def is_fd(self, value: bool) -> None:
        self._is_fd = value

    @property
    def name(self) -> str:
        """The message name as a string.

        """

        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def length(self) -> int:
        """The message data length in bytes.

        """

        return self._length

    @length.setter
    def length(self, value: int) -> None:
        self._length = value

    @property
    def signals(self) -> list[Signal]:
        """A list of all signals in the message.

        """

        return self._signals

    @property
    def is_container(self) -> bool:
        """Returns if the message is a container message

        """

        return len(self._contained_messages) > 0

    @property
    def contained_messages(self) -> list['Message']:
        """The list of messages potentially contained within this message

        """

        return self._contained_messages

    @property
    def unused_bit_pattern(self) -> int:
        """The pattern used for unused bits of a message.

        This prevents undefined behaviour and/or information leaks
        when encoding messages.
        """

        return self._unused_bit_pattern

    @unused_bit_pattern.setter
    def unused_bit_pattern(self, value: int) -> None:
        if value < 0 or value > 255:
            LOGGER.info(f'Invalid unused bit pattern "{value}". Must be '
                        f'an integer between 0 and 255')
            self._unused_bit_pattern = 0
            return

        self._unused_bit_pattern = value

    @property
    def signal_groups(self) -> list[SignalGroup] | None:
        """A list of all signal groups in the message.

        """

        return self._signal_groups

    @signal_groups.setter
    def signal_groups(self, value: list[SignalGroup]) -> None:
        self._signal_groups = value

    @property
    def comment(self) -> str | None:
        """The message comment, or ``None`` if unavailable.

        Note that we implicitly try to return the English comment if
        multiple languages were specified.

        """
        if self._comments is None:
            return None
        elif self._comments.get(None) is not None:
            return self._comments.get(None)
        elif self._comments.get('FOR-ALL') is not None:
            return self._comments.get('FOR-ALL')

        return self._comments.get('EN')

    @comment.setter
    def comment(self, value: str | None) -> None:
        if value is None:
            self._comments = None
        else:
            self._comments = Comments({None: value})

    @property
    def comments(self) -> Comments | None:
        """The dictionary with the descriptions of the message in multiple
        languages. ``None`` if unavailable.

        """
        return self._comments

    @comments.setter
    def comments(self, value: Comments) -> None:
        self._comments = value

    @property
    def senders(self) -> list[str]:
        """A list of all sender nodes of this message.

        """

        return self._senders

    @property
    def receivers(self) -> set[str]:
        """A set of all receiver nodes of this message.

        This is equivalent to the set of nodes which receive at least
        one of the signals contained in the message.

        """
        result: set[str] = set()

        for sig in self.signals:
            result.update(sig.receivers)

        if self.is_container:
            for cmsg in self.contained_messages:
                for sig in cmsg.signals:
                    result.update(sig.receivers)

        return result

    @property
    def send_type(self) -> str | None:
        """The message send type, or ``None`` if unavailable.

        """

        return self._send_type

    @property
    def cycle_time(self) -> int | None:
        """The message cycle time, or ``None`` if unavailable.

        """

        return self._cycle_time

    @cycle_time.setter
    def cycle_time(self, value: int | None) -> None:
        self._cycle_time = value

    @property
    def dbc(self) -> DbcSpecifics | None:
        """An object containing dbc specific properties like e.g. attributes.

        """

        return self._dbc

    @dbc.setter
    def dbc(self, value: DbcSpecifics | None) -> None:
        self._dbc = value

    @property
    def autosar(self) -> AutosarMessageSpecifics | None:
        """An object containing AUTOSAR specific properties

        e.g. auxiliary data required to implement CRCs, secure on-board
        communication (secOC) or container messages.
        """

        return self._autosar

    @autosar.setter
    def autosar(self, value: AutosarMessageSpecifics | None) -> None:
        self._autosar = value

    @property
    def bus_name(self) -> str | None:
        """The message bus name, or ``None`` if unavailable.

        """

        return self._bus_name

    @bus_name.setter
    def bus_name(self, value: str | None) -> None:
        self._bus_name = value

    @property
    def protocol(self) -> str | None:
        """The message protocol, or ``None`` if unavailable. Only one protocol
        is currently supported; ``'j1939'``.

        """

        return self._protocol

    @protocol.setter
    def protocol(self, value: str | None) -> None:
        self._protocol = value

    @property
    def signal_tree(self) -> SignalTreeType:
        """All signal names and multiplexer ids as a tree. Multiplexer signals
        are dictionaries, while other signals are strings.

        >>> foo = db.get_message_by_name('Foo')
        >>> foo.signal_tree
        ['Bar', 'Fum']
        >>> bar = db.get_message_by_name('Bar')
        >>> bar.signal_tree
        [{'A': {0: ['C', 'D'], 1: ['E']}}, 'B']

        """

        return self._signal_tree

    def gather_signals(self,
                       input_data: SignalMappingType,
                       node: Codec | None = None) \
      -> SignalDictType:

        '''Given a superset of all signals required to encode the message,
        return a dictionary containing exactly the ones required.

        If a required signal is missing from the input dictionary, a
        ``EncodeError`` exception is raised.
        '''

        if node is None:
            node = self._codecs
        assert node is not None

        result: dict[str, SignalValueType] = {}

        for signal in node['signals']:
            val = input_data.get(signal.name)
            if val is None:
                raise EncodeError(f'The signal "{signal.name}" is '
                                  f'required for encoding.')
            result[signal.name] = val

        for mux_signal_name, mux_nodes in node['multiplexers'].items():
            mux_num = self._get_mux_number(input_data, mux_signal_name)
            mux_node = mux_nodes.get(mux_num)
            if mux_node is None:
                multiplexers = node['multiplexers']
                try:
                    expected_str = \
                        f'Expected one of {{' \
                        f'{format_or(list(multiplexers[mux_signal_name].keys()))}' \
                        f'}}, but '
                except KeyError:
                    expected_str = ''

                raise EncodeError(f'A valid value for the multiplexer selector '
                                  f'signal "{mux_signal_name}" is required: '
                                  f'{expected_str}'
                                  f'got {input_data[mux_signal_name]}')

            result.update(self.gather_signals(input_data, mux_node))

        return result

    def gather_container(self,
                         contained_messages: list[ContainerHeaderSpecType],
                         signal_values: SignalMappingType) \
      -> ContainerDecodeResultType:

        '''Given a superset of all messages required to encode all messages
        featured by a container message, return a list of (Message,
        SignalDict) tuples that can be passed to ``encode()``.

        If a required signal is missing from the input dictionary, a
        ``EncodeError`` exception is raised.
        '''

        result: ContainerDecodeResultListType = []
        for header in contained_messages:
            contained_message = None
            if isinstance(header, str):
                contained_message = \
                    self.get_contained_message_by_name(header)
            elif isinstance(header, Message):
                # contained message is specified directly. We go once
                # around the circle to ensure that a contained message
                # with the given header ID is there.
                header_id = header.header_id
                assert header_id is not None
                contained_message = \
                    self.get_contained_message_by_header_id(header_id)
            elif isinstance(header, int):
                # contained message is specified directly. We go once
                # around the circle to ensure that a contained message
                # with the given header ID is there.
                contained_message = \
                    self.get_contained_message_by_header_id(header)

            if contained_message is None:
                raise EncodeError(f'Cannot determine contained message '
                                  f'associated with "{header}"')

            contained_signals = contained_message.gather_signals(signal_values)

            result.append( (contained_message, contained_signals) )

        return result

    def assert_signals_encodable(self,
                                 input_data: SignalMappingType,
                                 scaling: bool,
                                 assert_values_valid: bool = True,
                                 assert_all_known: bool = True) \
      -> None:

        '''Given a dictionary of signal name to signal value mappings, ensure
        that all the signals required for encoding are present

        As a minimum, all signals required to encode the message need
        to be specified. If they are not, a ``KeyError`` or an
        ``EncodeError`` exception is raised.

        Depending on the parameters specified, the data of the
        dictionary must adhere to additional requirements:

        :param scaling: If ``False`` no scaling of signals is performed.

        :param assert_values_valid: If ``True``, the values of all
            specified signals must be valid/encodable. If at least one is
            not, an ``EncodeError`` exception is raised. (Note that the
            values of multiplexer selector signals must always be valid!)

        :param assert_all_known: If ``True``, all specified signals must
            be used by the encoding operation or an ``EncodeError``
            exception is raised. This is useful to prevent typos.
        '''

        # this method only deals with ordinary messages
        if self.is_container:
            raise EncodeError(f'Message "{self.name}" is a container')

        # This type checking is not really comprehensive and is
        # superfluous if the type hints are respected by the calling
        # code. That said, it guards against accidentally passing
        # non-dictionary objects such as lists of (Message,
        # SignalDict) tuples expected by container messages...
        if not isinstance(input_data, dict):
            raise EncodeError(f'Input data for encoding message "{self.name}" '
                              f'must be a SignalDict')

        used_signals = self.gather_signals(input_data)
        if assert_all_known and set(used_signals) != set(input_data):
            raise EncodeError(f'The following signals were specified but are '
                              f'not required to encode the message:'
                              f'{set(input_data) - set(used_signals)}')
        if assert_values_valid:
            self._assert_signal_values_valid(used_signals, scaling)

    def assert_container_encodable(self,
                                   input_data: ContainerEncodeInputType,
                                   scaling: bool,
                                   assert_values_valid: bool = True,
                                   assert_all_known: bool = True) \
      -> None:

        """
        This method is identical to ``assert_signals_encodable()``
        except that it is concerned with container messages.
        """

        # this method only deals with container messages
        if not self.is_container:
            raise EncodeError(f'Message "{self.name}" is not a container')

        # This type checking is not really comprehensive and is
        # superfluous if the type hints are respected by the calling
        # code. That said it guards against accidentally passing a
        # SignalDict for normal messages...
        if not isinstance(input_data, list):
            raise EncodeError(f'Input data for encoding message "{self.name}" '
                              f'must be a list of (Message, SignalDict) tuples')

        for header, payload in input_data:
            if isinstance(header, int) and isinstance(payload, bytes):
                # contained message specified as raw data
                continue

            contained_message = None
            if isinstance(header, int):
                contained_message = \
                    self.get_contained_message_by_header_id(header)
            elif isinstance(header, str):
                contained_message = \
                    self.get_contained_message_by_name(header)
            elif isinstance(header, Message):
                hid = header.header_id
                if hid is None:
                    raise EncodeError(f'Message {header.name} cannot be part '
                                      f'of a container because it does not '
                                      f'exhibit a header ID')
                contained_message = self.get_contained_message_by_header_id(hid)

            if contained_message is None:
                raise EncodeError(f'Could not associate "{header}" with any '
                                  f'contained message')

            if isinstance(payload, bytes):
                if len(payload) != contained_message.length:
                    raise EncodeError(f'Payload for contained message '
                                      f'"{contained_message.name}" is '
                                      f'{len(payload)} instead of '
                                      f'{contained_message.length} bytes long')
            else:
                contained_message.assert_signals_encodable(payload,
                                                           scaling,
                                                           assert_values_valid,
                                                           assert_all_known)

    def _get_mux_number(self, decoded: SignalMappingType, signal_name: str) -> int:
        mux = decoded[signal_name]

        if isinstance(mux, (str, NamedSignalValue)):
            signal = self.get_signal_by_name(signal_name)
            try:
                mux = signal.conversion.choice_to_number(str(mux))
            except KeyError:
                raise EncodeError() from None
        return int(mux)

    def _assert_signal_values_valid(self,
                                    data: SignalMappingType,
                                    scaling: bool) -> None:

        for signal_name, signal_value in data.items():
            signal = self.get_signal_by_name(signal_name)

            if isinstance(signal_value, (str, NamedSignalValue)):
                # Check choices
                signal_value_num = signal.conversion.choice_to_number(str(signal_value))

                if signal_value_num is None:
                    raise EncodeError(f'Invalid value specified for signal '
                                      f'"{signal.name}": "{signal_value}"')
                continue

            # retrieve the signal's scaled value to perform range check against minimum and maximum,
            # retrieve the signal's raw value to check if exists in value table
            if scaling:
                scaled_value = signal_value
                raw_value = signal.conversion.numeric_scaled_to_raw(scaled_value)
            else:
                scaled_value = cast(
                    'int | float',
                    signal.conversion.raw_to_scaled(raw_value=signal_value, decode_choices=False)
                )
                raw_value = signal_value

            if signal.conversion.choices and raw_value in signal.conversion.choices:
                # skip range check if raw value exists in value table
                continue

            if signal.minimum is not None and scaled_value < signal.minimum - abs(signal.conversion.scale)*1e-6:
                    raise EncodeError(
                        f'Expected signal "{signal.name}" value greater than '
                        f'or equal to {signal.minimum} in message "{self.name}", '
                        f'but got {scaled_value}.')

            if signal.maximum is not None and scaled_value > signal.maximum + abs(signal.conversion.scale)*1e-6:
                raise EncodeError(
                    f'Expected signal "{signal.name}" value smaller than '
                    f'or equal to {signal.maximum} in message "{self.name}", '
                    f'but got {scaled_value}.')

    def _encode(self, node: Codec, data: SignalMappingType, scaling: bool) -> tuple[int, int, list[Signal]]:
        encoded = encode_data(data,
                              node['signals'],
                              node['formats'],
                              scaling)
        padding_mask = node['formats'].padding_mask
        multiplexers = node['multiplexers']

        all_signals = list(node['signals'])
        for signal in multiplexers:
            mux = self._get_mux_number(data, signal)

            try:
                node = multiplexers[signal][mux]
            except KeyError:
                raise EncodeError(f'Expected multiplexer id in '
                                  f'{{{format_or(list(multiplexers[signal].keys()))}}}, '
                                  f'for multiplexer "{signal}" '
                                  f'but got {mux}') from None

            mux_encoded, mux_padding_mask, mux_signals = \
                self._encode(node, data, scaling)
            all_signals.extend(mux_signals)

            encoded |= mux_encoded
            padding_mask &= mux_padding_mask

        return encoded, padding_mask, all_signals

    def _encode_container(self,
                          data: ContainerEncodeInputType,
                          scaling: bool,
                          padding: bool) -> bytes:

        result = b""

        for header, value in data:
            if isinstance(header, str):
                contained_message = \
                    self.get_contained_message_by_name(header)
            elif isinstance(header, Message):
                # contained message is specified directly. We go once
                # around the circle to ensure that a contained message
                # with the given header ID is there.
                contained_message = \
                    self.get_contained_message_by_header_id(header.header_id) # type: ignore
            elif isinstance(header, int):
                # contained message is specified directly. We go once
                # around the circle to ensure that a contained message
                # with the given header ID is there.
                contained_message = \
                    self.get_contained_message_by_header_id(header)
            else:
                raise EncodeError(f'Could not determine message corresponding '
                                  f'to header {header}')

            if contained_message is None:
                if isinstance(value, bytes) and isinstance(header, int):
                    # the contained message was specified as raw data
                    header_id = header
                else:
                    raise EncodeError(f'No message corresponding to header '
                                      f'{header} could be determined')
            else:
                assert contained_message.header_id is not None
                header_id = contained_message.header_id

            if isinstance(value, bytes):
                # raw data

                # produce a message if size of the blob does not
                # correspond to the size specified by the message
                # which it represents.
                if contained_message is not None and \
                    len(value) != contained_message.length:

                    LOGGER.info(f'Specified data for contained message '
                                f'{contained_message.name} is '
                                f'{len(value)} bytes instead of '
                                f'{contained_message.length} bytes')

                contained_payload = value

            elif isinstance(value, dict):
                # signal_name to signal_value dictionary
                assert contained_message is not None
                contained_payload = contained_message.encode(value,
                                                             scaling,
                                                             padding,
                                                             strict=False)

            else:
                assert contained_message is not None
                raise EncodeError(f'Cannot encode payload for contained '
                                  f'message "{contained_message.name}".')

            hbo = 'big' if self.header_byte_order == 'big_endian' else 'little'
            result += int.to_bytes(header_id,
                                   3,
                                   hbo) # type: ignore
            result += int.to_bytes(len(contained_payload), 1, 'big')
            result += bytes(contained_payload)

        return result

    def encode(self,
               data: EncodeInputType,
               scaling: bool = True,
               padding: bool = False,
               strict: bool = True,
               ) -> bytes:

        """Encode given data as a message of this type.

        If the message is an "ordinary" frame, this method expects a
        key-to-value dictionary as `data` which maps the name of every
        required signal to a value that can be encoded by that
        signal. If the current message is a container message, it
        expects a list of `(contained_message, contained_data)` tuples
        where `contained_message` is either an integer with the header
        ID, the name or the message object of the contained
        message. Similarly, the `contained_data` can either be
        specified as raw binary data (`bytes`) or as a key-to-value
        dictionary of every signal needed to encode the featured
        message.

        If `scaling` is ``False`` no scaling of signals is performed.

        If `padding` is ``True`` unused bits are encoded as 1.

        If `strict` is ``True`` the specified signals must exactly be the
        ones expected, and their values must be within their allowed ranges,
        or an `EncodeError` exception is raised.

        >>> foo = db.get_message_by_name('Foo')
        >>> foo.encode({'Bar': 1, 'Fum': 5.0})
        b'\\x01\\x45\\x23\\x00\\x11'

        """

        if self.is_container:
            if strict:
                if not isinstance(data, (list, tuple)):
                    raise EncodeError('Container frames can only encode lists of '
                                      '(message, data) tuples')

                self.assert_container_encodable(data, scaling=scaling)

            return self._encode_container(cast('ContainerEncodeInputType', data),
                                          scaling,
                                          padding)

        if strict:
            # setting 'strict' to True is just a shortcut for calling
            # 'assert_signals_encodable()' using the strictest
            # settings.
            if not isinstance(data, dict):
                raise EncodeError('The payload for encoding non-container '
                                  'messages must be a signal name to '
                                  'signal value dictionary')
            self.assert_signals_encodable(data, scaling=scaling)

        if self._codecs is None:
            raise ValueError('Codec is not initialized.')

        encoded, padding_mask, _ = self._encode(self._codecs,
                                                cast('SignalMappingType', data),
                                                scaling)

        if padding:
            padding_pattern = int.from_bytes([self._unused_bit_pattern] * self._length, "big")
            encoded |= (padding_mask & padding_pattern)

        return encoded.to_bytes(self._length, "big")

    def _decode(self,
                node: Codec,
                data: bytes,
                decode_choices: bool,
                scaling: bool,
                allow_truncated: bool,
                allow_excess: bool) -> SignalDictType:
        decoded = decode_data(data,
                              self.length,
                              node['signals'],
                              node['formats'],
                              decode_choices,
                              scaling,
                              allow_truncated,
                              allow_excess)

        multiplexers = node['multiplexers']

        for signal in multiplexers:
            if allow_truncated and signal not in decoded:
                continue

            mux = self._get_mux_number(decoded, signal)

            try:
                node = multiplexers[signal][mux]
            except KeyError:
                raise DecodeError(f'expected multiplexer id {format_or(sorted(multiplexers[signal].keys()))}, but got {mux}') from None

            decoded.update(self._decode(node,
                                        data,
                                        decode_choices,
                                        scaling,
                                        allow_truncated,
                                        allow_excess))

        return decoded

    def unpack_container(self,
                         data: bytes,
                         allow_truncated: bool = False) \
                         -> ContainerUnpackResultType:
        """Unwrap the contents of a container message.

        This returns a list of ``(contained_message, contained_data)``
        tuples, i.e., the data for the contained message are ``bytes``
        objects, not decoded signal dictionaries. This is required for
        verifying the correctness of the end-to-end protection or the
        authenticity of a contained message.

        Note that ``contained_message`` is the header ID integer value
        if a contained message is unknown. Further, if something goes
        seriously wrong, a ``DecodeError`` is raised.
        """

        if not self.is_container:
            raise DecodeError(f'Cannot unpack non-container message '
                              f'"{self.name}"')

        if len(data) > self.length:
            raise DecodeError(f'Container message "{self.name}" specified '
                              f'as exhibiting at most {self.length} but '
                              f'received a {len(data)} bytes long frame')

        result: ContainerUnpackListType = []
        pos = 0
        while pos < len(data):
            if pos + 4 > len(data):
                # TODO: better throw an exception? only warn in strict mode?
                LOGGER.info(f'Malformed container message '
                            f'"{self.name}" encountered while decoding: '
                            f'No valid header specified for contained '
                            f'message #{len(result)+1} starting at position '
                            f'{pos}. Ignoring.')
                return result

            contained_id = int.from_bytes(data[pos:pos+3], 'big')
            contained_len = data[pos+3]

            if pos + 4 + contained_len > len(data):
                if not allow_truncated:
                    raise DecodeError(f'Malformed container message '
                                      f'"{self.name}": Contained message '
                                      f'{len(result)+1} would exceed total '
                                      f'message size.')
                else:
                    contained_len = len(data) - pos - 4


            contained_data = data[pos+4:pos+4+contained_len]
            contained_msg = \
                self.get_contained_message_by_header_id(contained_id)
            pos += 4+contained_len

            if contained_msg is None:
                result.append((contained_id, bytes(contained_data)))
            else:
                result.append((contained_msg, bytes(contained_data)))

        return result

    def decode(self,
               data: bytes,
               decode_choices: bool = True,
               scaling: bool = True,
               decode_containers: bool = False,
               allow_truncated: bool = False,
               allow_excess: bool = True,
               ) \
               -> DecodeResultType:
        """Decode given data as a message of this type.

        If `decode_choices` is ``False`` scaled values are not
        converted to choice strings (if available).

        If `scaling` is ``False`` no scaling of signals is performed.

        >>> foo = db.get_message_by_name('Foo')
        >>> foo.decode(b'\\x01\\x45\\x23\\x00\\x11')
        {'Bar': 1, 'Fum': 5.0}

        If `decode_containers` is ``True``, the inner messages are
        decoded if the current message is a container frame. The
        reason why this needs to be explicitly enabled is that the
        result of `decode()` for container frames is a list of
        ``(header_id, signals_dict)`` tuples which might cause code
        that does not expect this to misbehave. Trying to decode a
        container message with `decode_containers` set to ``False``
        will raise a `DecodeError`.

        If `allow_truncated` is ``True``, incomplete messages (i.e.,
        ones where the received data is shorter than specified) will
        be partially decoded, i.e., all signals which are fully
        present in the received data will be decoded, and the
        remaining ones will be omitted. If 'allow_truncated` is set to
        ``False``, `DecodeError` will be raised when trying to decode
        incomplete messages.

        If `allow_excess` is ``True``, data that is are longer than
        the expected message length is decoded, else a `ValueError` is
        raised if such data is encountered.
        """

        if decode_containers and self.is_container:
            return self.decode_container(data,
                                         decode_choices,
                                         scaling,
                                         allow_truncated,
                                         allow_excess)

        return self.decode_simple(data,
                                  decode_choices,
                                  scaling,
                                  allow_truncated,
                                  allow_excess)

    def decode_simple(self,
                      data: bytes,
                      decode_choices: bool = True,
                      scaling: bool = True,
                      allow_truncated: bool = False,
                      allow_excess: bool = True) \
                      -> SignalDictType:
        """Decode given data as a container message.

        This method is identical to ``decode()`` except that the
        message **must not** be a container. If the message is a
        container, an exception is raised.
        """

        if self.is_container:
            raise DecodeError(f'Message "{self.name}" is a container')
        elif self._codecs is None:
            raise ValueError('Codec is not initialized.')

        return self._decode(self._codecs,
                            data,
                            decode_choices,
                            scaling,
                            allow_truncated,
                            allow_excess)

    def decode_container(self,
                         data: bytes,
                         decode_choices: bool = True,
                         scaling: bool = True,
                         allow_truncated: bool = False,
                         allow_excess: bool = True) \
                         -> ContainerDecodeResultType:
        """Decode given data as a container message.

        This method is identical to ``decode()`` except that the
        message **must** be a container. If the message is not a
        container, an exception is raised.
        """

        if not self.is_container:
            raise DecodeError(f'Message "{self.name}" is not a container')

        unpacked = self.unpack_container(data, allow_truncated)

        result: ContainerDecodeResultListType = []

        for contained_message, contained_data in unpacked:
            if not isinstance(contained_message, Message):
                result.append((contained_message, bytes(contained_data)))
                continue

            try:
                decoded = contained_message.decode(contained_data,
                                                   decode_choices,
                                                   scaling,
                                                   decode_containers=False,
                                                   allow_truncated=allow_truncated,
                                                   allow_excess=allow_excess)
            except (ValueError, DecodeError):
                result.append((contained_message, bytes(contained_data)))
                continue

            result.append((contained_message, decoded)) # type: ignore

        return result

    def get_contained_message_by_header_id(self, header_id: int) \
        -> Optional['Message']:

        tmp = [ x for x in self.contained_messages if x.header_id == header_id ]

        if len(tmp) == 0:
            return None
        elif len(tmp) > 1:
            raise Error(f'Container message "{self.name}" contains multiple '
                        f'contained messages exhibiting id 0x{header_id:x}')

        return tmp[0]

    def get_contained_message_by_name(self, name: str) \
        -> Optional['Message']:

        tmp = [ x for x in self.contained_messages if x.name == name ]

        if len(tmp) == 0:
            return None
        elif len(tmp) > 1:
            raise Error(f'Container message "{self.name}" contains multiple '
                        f'contained messages named "{name}"')

        return tmp[0]

    def get_signal_by_name(self, name: str) -> Signal:
        return self._signal_dict[name]

    def is_multiplexed(self) -> bool:
        """Returns ``True`` if the message is multiplexed, otherwise
        ``False``.

        >>> foo = db.get_message_by_name('Foo')
        >>> foo.is_multiplexed()
        False
        >>> bar = db.get_message_by_name('Bar')
        >>> bar.is_multiplexed()
        True

        """
        if self._codecs is None:
            raise ValueError('Codec is not initialized.')

        return bool(self._codecs['multiplexers'])

    def _check_signal(self, message_bits: MutableSequence[str | None], signal: Signal) -> None:
        signal_bits: list[str | None] = cast('list[str | None]', signal.length * [signal.name])

        if signal.byte_order == 'big_endian':
            padding = cast('list[str | None]', start_bit(signal) * [None])
            signal_bits = padding + signal_bits
        else:
            signal_bits = signal_bits + cast('list[str | None]', signal.start * [None])

            if len(signal_bits) < len(message_bits):
                padding = cast('list[str | None]', (len(message_bits) - len(signal_bits)) * [None])
                reversed_signal_bits = padding + signal_bits
            else:
                reversed_signal_bits = signal_bits

            signal_bits = []

            for i in range(0, len(reversed_signal_bits), 8):
                signal_bits = reversed_signal_bits[i:i + 8] + signal_bits

        # Check that the signal fits in the message.
        if len(signal_bits) > len(message_bits):
            raise Error(f'The signal {signal.name} does not fit in message {self.name}.')

        # Check that the signal does not overlap with other
        # signals.
        for offset, signal_bit in enumerate(signal_bits):
            if signal_bit is not None:
                if message_bits[offset] is not None:
                    raise Error(
                        f'The signals {signal.name} and {message_bits[offset]} are overlapping in message {self.name}.')

                message_bits[offset] = signal.name

    def _check_mux(self, message_bits: MutableSequence[str | None], mux: dict[str, dict[int, Sequence[str]]]) -> None:
        signal_name, children = next(iter(mux.items()))
        self._check_signal(message_bits,
                           self.get_signal_by_name(signal_name))
        children_message_bits = deepcopy(message_bits)

        for multiplexer_id in sorted(children):
            child_tree = children[multiplexer_id]
            child_message_bits = deepcopy(children_message_bits)
            self._check_signal_tree(child_message_bits, child_tree)

            for i, child_bit in enumerate(child_message_bits):
                if child_bit is not None:
                    message_bits[i] = child_bit

    def _check_signal_tree(self, message_bits: MutableSequence[str | None], signal_tree: SignalTreeType) -> None:
        for signal_name in signal_tree:
            if isinstance(signal_name, dict):
                self._check_mux(message_bits, signal_name)
            else:
                self._check_signal(message_bits,
                                   self.get_signal_by_name(signal_name))

    def _check_signal_lengths(self) -> None:
        for signal in self._signals:
            if signal.length <= 0:
                raise Error(
                    f'The signal {signal.name} length {signal.length} is not greater than 0 in '
                    f'message {self.name}.')

    def refresh(self, strict: bool | None = None) -> None:
        """Refresh the internal message state.

        If `strict` is ``True`` an exception is raised if any signals
        are overlapping or if they don't fit in the message. This
        argument overrides the value of the same argument passed to
        the constructor.

        """

        self._check_signal_lengths()
        self._codecs = self._create_codec()
        self._signal_tree = self._create_signal_tree(self._codecs)
        self._signal_dict = {signal.name: signal for signal in self._signals}

        if strict is None:
            strict = self._strict

        if strict:
            message_bits: MutableSequence[str | None] = cast(
                'MutableSequence[str | None]',
                [None] * (8 * self.length)
            )
            self._check_signal_tree(message_bits, self._signal_tree)

    def __repr__(self) -> str:
        return \
            f'message(' \
            f"'{self._name}', " \
            f'0x{self._frame_id:x}, ' \
            f'{self._is_extended_frame}, '\
            f'{self._length}, ' \
            f'{self._comments})'



#!/usr/bin/env python3
#
# Script testing DBC manipulation and io.
#

import os

import cantools

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
INPUT_DBC_PATH = os.path.join(SCRIPT_DIR, 'dbc_input.dbc')
OUTPUT_DBC_PATH = os.path.join(SCRIPT_DIR, 'dbc_output.dbc')

# Read the DBC.
print(f"Loading DBC database from '{INPUT_DBC_PATH}'.")
db = cantools.db.load_file(INPUT_DBC_PATH)

# Get a message to manipulate.
message = db.get_message_by_name('TheMessage')

# Manipulate a message attribute. Try/except will catch accesses to
# non existent attributes.
try:
    print("Input cycle time: ", message.cycle_time)
    message.cycle_time = 2000
    print("Output cycle time:", message.cycle_time)
except KeyError:
    raise

# Manipulate the message frame id.
print("Input frame id: ", hex(message.frame_id))
message.frame_id = 0x234
print("Output frame id:", hex(message.frame_id))

print(f"Writing modified DBC database to '{OUTPUT_DBC_PATH}'.")

with open(OUTPUT_DBC_PATH, 'w', newline="\r\n") as fout:
    fout.write(db.as_dbc_string())



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union

from ..typechecking import Choices, SignalValueType

if TYPE_CHECKING:
    from .namedsignalvalue import NamedSignalValue


class BaseConversion(ABC):
    """The BaseConversion class defines the interface for all signal conversion classes."""

    #: the scaling factor of the conversion
    scale: float

    #: the offset value of the conversion
    offset: float

    #: ``True`` if the raw/internal value is a floating datatype
    #: ``False`` if it is an integer datatype
    is_float: bool

    #: an optional mapping of raw values to their corresponding text value
    choices: Choices | None

    @staticmethod
    def factory(
        scale: float = 1,
        offset: float = 0,
        choices: Choices | None = None,
        is_float: bool = False,
    ) -> "BaseConversion":
        """Factory method that returns an instance of a conversion subclass based on the given parameters.

        :param scale:
            The scale factor to use for the conversion.
        :param offset:
            The offset to use for the conversion.
        :param choices:
            A dictionary of named signal choices, mapping raw values to string labels.
        :param is_float:
            A boolean flag indicating whether the raw value is a float or an integer.

        :returns:
            An instance of a conversion subclass, either an `IdentityConversion`, a `LinearIntegerConversion`,
            a `LinearConversion`or a `NamedSignalConversion`.

        :raises TypeError: If the given parameters are of the wrong type.
        """
        if choices is None:
            if scale == 1 and offset == 0:
                return IdentityConversion(is_float=is_float)

            if _is_integer(scale) and _is_integer(offset) and not is_float:
                return LinearIntegerConversion(scale=int(scale), offset=int(offset))

            return LinearConversion(
                scale=scale,
                offset=offset,
                is_float=is_float,
            )

        return NamedSignalConversion(
            scale=scale, offset=offset, choices=choices, is_float=is_float
        )

    @abstractmethod
    def raw_to_scaled(
        self,
        raw_value: float,
        decode_choices: bool = True,
    ) -> SignalValueType:
        """Convert an internal raw value according to the defined scaling or value table.

        :param raw_value:
            The raw value
        :param decode_choices:
            If `decode_choices` is ``False`` scaled values are not
            converted to choice strings (if available).
        :return:
            The calculated scaled value
        """
        raise NotImplementedError

    @abstractmethod
    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        """Convert a scaled value to the internal raw value.

        :param scaled_value:
            The scaled value.
        :return:
            The internal raw value.
        """
        raise NotImplementedError

    @abstractmethod
    def numeric_scaled_to_raw(
        self, scaled_value: float
    ) -> int | float:
        """Convert a numeric scaled value to the internal raw value.

        :param scaled_value:
            The numeric scaled value.
        :return:
            The internal raw value.
        """
        raise NotImplementedError

    def choice_to_number(self, choice: Union[str, "NamedSignalValue"]) -> int:
        raise KeyError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


class IdentityConversion(BaseConversion):
    scale = 1
    offset = 0
    choices = None

    def __init__(self, is_float: bool) -> None:
        self.is_float = is_float

    def raw_to_scaled(
        self,
        raw_value: float,
        decode_choices: bool = True,
    ) -> int | float:
        return raw_value

    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        if not isinstance(scaled_value, (int, float)):
            raise TypeError(
                f"'scaled_value' must have type 'int' or 'float' (is {type(scaled_value)})"
            )
        return self.numeric_scaled_to_raw(scaled_value)

    def numeric_scaled_to_raw(
        self, scaled_value: float
    ) -> int | float:
        return scaled_value if self.is_float else round(scaled_value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(is_float={self.is_float})"


class LinearIntegerConversion(BaseConversion):
    is_float = False
    choices = None

    def __init__(self, scale: int, offset: int) -> None:
        self.scale: int = scale
        self.offset: int = offset

    def raw_to_scaled(
        self,
        raw_value: float,
        decode_choices: bool = True,
    ) -> SignalValueType:
        return raw_value * self.scale + self.offset

    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        if not isinstance(scaled_value, (int, float)):
            raise TypeError(
                f"'scaled_value' must have type 'int' or 'float' (is {type(scaled_value)})"
            )
        return self.numeric_scaled_to_raw(scaled_value)

    def numeric_scaled_to_raw(
        self, scaled_value: float
    ) -> int | float:
        # try to avoid a loss of precision whenever possible
        _raw = scaled_value - self.offset
        quotient, remainder = divmod(_raw, self.scale)
        if remainder == 0:
            _raw = quotient
        else:
            _raw /= self.scale
        return round(_raw)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(scale={self.scale}, offset={self.offset})"


class LinearConversion(BaseConversion):
    choices = None

    def __init__(self, scale: float, offset: float, is_float: bool) -> None:
        self.scale = scale
        self.offset = offset
        self.is_float = is_float

    def raw_to_scaled(
        self,
        raw_value: float,
        decode_choices: bool = True,
    ) -> SignalValueType:
        return raw_value * self.scale + self.offset

    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        if not isinstance(scaled_value, (int, float)):
            raise TypeError(
                f"'scaled_value' must have type 'int' or 'float' (is {type(scaled_value)})"
            )
        return self.numeric_scaled_to_raw(scaled_value)

    def numeric_scaled_to_raw(
        self, scaled_value: float
    ) -> int | float:
        _raw = (scaled_value - self.offset) / self.scale
        return _raw if self.is_float else round(_raw)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"scale={self.scale}, "
            f"offset={self.offset}, "
            f"is_float={self.is_float})"
        )


class NamedSignalConversion(BaseConversion):
    def __init__(
        self, scale: float, offset: float, choices: Choices, is_float: bool
    ) -> None:
        self.scale = scale
        self.offset = offset
        self.is_float = is_float
        self._inverse_choices: dict[str, int] = {}
        self.choices: Choices = choices
        self._update_choices()

        self._conversion = BaseConversion.factory(
            scale=self.scale,
            offset=self.offset,
            choices=None,
            is_float=is_float,
        )
        # monkeypatch method to avoid unnecessary function call
        self.numeric_scaled_to_raw = self._conversion.numeric_scaled_to_raw  # type: ignore[method-assign]

    def raw_to_scaled(
        self,
        raw_value: float,
        decode_choices: bool = True,
    ) -> SignalValueType:
        if decode_choices and (choice := self.choices.get(int(raw_value))) is not None:
            return choice
        return self._conversion.raw_to_scaled(raw_value, False)

    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        if isinstance(scaled_value, (int, float)):
            return self._conversion.scaled_to_raw(scaled_value)

        if hasattr(scaled_value, "value"):
            # scaled_value is NamedSignalValue
            return scaled_value.value

        if isinstance(scaled_value, str):
            return self.choice_to_number(scaled_value)

        raise TypeError

    def numeric_scaled_to_raw(
        self, scaled_value: float
    ) -> int | float:
        return self._conversion.scaled_to_raw(scaled_value)

    def set_choices(self, choices: Choices) -> None:
        self.choices = choices
        self._update_choices()

    def _update_choices(self) -> None:
        # we simply assume that the choices are invertible
        self._inverse_choices = {str(x[1]): x[0] for x in self.choices.items()}

    def choice_to_number(self, choice: Union[str, "NamedSignalValue"]) -> int:
        return self._inverse_choices[str(choice)]

    def __repr__(self) -> str:
        list_of_choices = ", ".join(
            [f"{value}: '{text}'" for value, text in self.choices.items()]
        )
        choices = f"{{{list_of_choices}}}"
        return (
            f"{self.__class__.__name__}("
            f"scale={self.scale}, "
            f"offset={self.offset}, "
            f"choices={choices}, "
            f"is_float={self.is_float})"
        )


def _is_integer(value: float) -> bool:
    if isinstance(value, int) or (hasattr(value, "is_integer") and value.is_integer()):
        return True
    elif isinstance(value, float):
        return False

    err_msg = f"`value` must be of type `int` or `float`, is {type(value)}"
    raise TypeError(err_msg)



# A signal group.


class SignalGroup:
    """A CAN signal group. Signal groups are used to define a group of
    signals within a message, e.g. to define that the signals of a
    group have to be updated in common.

    """

    def __init__(self,
                 name: str,
                 repetitions: int = 1,
                 signal_names: list[str] | None = None) -> None:
        self._name = name
        self._repetitions = repetitions
        self._signal_names = signal_names or []

    @property
    def name(self) -> str:
        """The signal group name as a string.

        """

        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def repetitions(self) -> int:
        """The signal group repetitions.

        """

        return self._repetitions

    @repetitions.setter
    def repetitions(self, value: int) -> None:
        self._repetitions = value

    @property
    def signal_names(self) -> list[str]:
        """The signal names in the signal group

        """

        return self._signal_names

    @signal_names.setter
    def signal_names(self, value: list[str]) -> None:
        self._signal_names = value

    def __repr__(self) -> str:
        return f"signal_group('{self._name}', {self._repetitions}, {self._signal_names})"



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cantools.database.can.formats.dbc_specifics import DbcSpecifics


class EnvironmentVariable:
    """A CAN environment variable.

    """

    def __init__(self,
                 name: str,
                 env_type: int,
                 minimum: float,
                 maximum: float,
                 unit: str,
                 initial_value: float,
                 env_id: int,
                 access_type: str,
                 access_node: str,
                 comment: str | None,
                 dbc_specifics: "DbcSpecifics") -> None:
        self._name = name
        self._env_type = env_type
        self._minimum = minimum
        self._maximum = maximum
        self._unit = unit
        self._initial_value = initial_value
        self._env_id = env_id
        self._access_type = access_type
        self._access_node = access_node
        self._comment = comment
        self._dbc = dbc_specifics

    @property
    def name(self) -> str:
        """The environment variable name as a string.

        """

        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def env_type(self) -> int:
        """The environment variable type value.

        """

        return self._env_type

    @env_type.setter
    def env_type(self, value: int) -> None:
        self._env_type = value

    @property
    def minimum(self) -> int | float:
        """The minimum value of the environment variable.

        """

        return self._minimum

    @minimum.setter
    def minimum(self, value: float) -> None:
        self._minimum = value

    @property
    def maximum(self) -> int | float:
        """The maximum value of the environment variable.

        """

        return self._maximum

    @maximum.setter
    def maximum(self, value: float) -> None:
        self._maximum = value

    @property
    def unit(self) -> str:
        """ The units in which the environment variable is expressed as a string.

        """

        return self._unit

    @unit.setter
    def unit(self, value: str) -> None:
        self._unit = value

    @property
    def initial_value(self) -> int | float:
        """The initial value of the environment variable.

        """

        return self._initial_value

    @initial_value.setter
    def initial_value(self, value: float) -> None:
        self._initial_value = value

    @property
    def env_id(self) -> int:
        """The id value of the environment variable.

        """

        return self._env_id

    @env_id.setter
    def env_id(self, value: int) -> None:
        self._env_id = value

    @property
    def access_type(self) -> str:
        """The environment variable access type as a string.

        """

        return self._access_type

    @access_type.setter
    def access_type(self, value: str) -> None:
        self._access_type = value

    @property
    def access_node(self) -> str:
        """The environment variable access node as a string.

        """

        return self._access_node

    @access_node.setter
    def access_node(self, value: str) -> None:
        self._access_node = value

    @property
    def comment(self) -> str | None:
        """The environment variable comment, or ``None`` if unavailable.

        """

        return self._comment

    @comment.setter
    def comment(self, value: str | None) -> None:
        self._comment = value

    @property
    def dbc(self) -> "DbcSpecifics":
        """An object containing dbc specific properties like e.g. attributes.

        """

        return self._dbc

    @dbc.setter
    def dbc(self, value: "DbcSpecifics") -> None:
        self._dbc = value

    def __repr__(self) -> str:
        return "environment_variable('{}', {}, {}, {}, '{}', {}, {}, '{}', '{}', {})".format(
            self._name,
            self._env_type,
            self._minimum,
            self._maximum,
            self._unit,
            self._initial_value,
            self._env_id,
            self._access_type,
            self._access_node,
            "'" + self._comment + "'" if self._comment is not None else None)



# An example cipher suite for secure on-board communication. This is
# in no way cryptographically secure. DO NOT USE IN THE REAL WORLD!


from ..database import Message


class SnakeOilAuthenticator:
    """A snake oil authenticator for secure on-board communication

    The sole purpose of this class is to demonstrate how SecOC can be
    implemented using cantools. These algorithms are in no way
    cryptographically secure! DO NOT USE THEM IN THE REAL WORLD!
    """
    def __init__(self,
                 secret: bytes | str) -> None:
        if isinstance(secret, str):
            self._secret = secret.encode()
        else:
            self._secret = bytes(secret)

    def __call__(self,
                 dbmsg: Message,
                 auth_data: bytearray,
                 freshness_value: int) \
                -> bytearray:

        v0 = freshness_value%253

        # XOR the secret and the data which we ought to authenticate
        result = bytearray([v0]*5)
        for i in range(len(auth_data)):
            result[i % len(result)] ^= auth_data[i]
            result[i % len(result)] ^= self._secret[i%len(self._secret)]

        return result



#!/usr/bin/env python3
#
# The "real" Motor node, transmitting the MotorStatus message
# periodically.
#

import struct

import can


def create_message(speed, load):
    return can.Message(arbitration_id=0x010,
                       is_extended_id=False,
                       data=struct.pack('<HB', speed, load))


def main():
    can.rc['interface'] = 'socketcan'
    can.rc['channel'] = 'vcan0'
    can_bus = can.interface.Bus()

    task = can_bus.send_periodic(create_message(0, 0), 1.0)

    while True:
        message = can_bus.recv()

        if message.arbitration_id == 0x011:
            speed = struct.unpack('<H', message.data)[0]
            print(f'Received motor speed of {speed} rpm.')
            task.modify_data(create_message(speed, 12))

            if message.data == b'\xff\xff':
                break

    print('Done!')


if __name__ == '__main__':
    main()





def num(number_as_string: str) -> int | float:
    """Convert given string to an integer or a float.

    """

    try:
        return int(number_as_string)
    except ValueError:
        return float(number_as_string)
    except Exception:
        raise ValueError('Expected integer or floating point number.') from None
